@@ -5,37 +5,40 @@ import logging
 import math
 import os
 import random
+import sqlite3 as sqlite
 import threading
 import time
 import urllib.parse
 import warnings
 from decimal import Decimal
+from sqlite3 import DatabaseError
+from sqlite3 import IntegrityError as DatabaseIntegrityError
+from sqlite3 import OperationalError as DatabaseOperationalError
 from weakref import WeakKeyDictionary
-from werkzeug.security import safe_join
 
-try:
-    from pysqlite2 import dbapi2 as sqlite
-    from pysqlite2.dbapi2 import IntegrityError as DatabaseIntegrityError
-    from pysqlite2.dbapi2 import OperationalError as DatabaseOperationalError
-except ImportError:
-    import sqlite3 as sqlite
-    from sqlite3 import IntegrityError as DatabaseIntegrityError
-    from sqlite3 import OperationalError as DatabaseOperationalError
-from sql import Flavor, Table, Query, Expression, Literal, Null
+from sql import Expression, Flavor, Literal, Null, Query, Table
 from sql.conditionals import NullIf
-from sql.functions import (Function, Extract, Position, Substring,
-    Overlay, CharLength, CurrentTimestamp, Trim)
+from sql.functions import (
+    CharLength, CurrentTimestamp, Extract, Function, Overlay, Position,
+    Substring, Trim)
 
 from trytond.backend.database import DatabaseInterface, SQLType
 from trytond.config import config, parse_uri
+from trytond.tools import safe_join
 from trytond.transaction import Transaction
 
-__all__ = ['Database', 'DatabaseIntegrityError', 'DatabaseOperationalError']
+__all__ = [
+    'Database',
+    'DatabaseIntegrityError', 'DatabaseDataError', 'DatabaseOperationalError']
 logger = logging.getLogger(__name__)
 
 _default_name = config.get('database', 'default_name', default=':memory:')
 
 
+class DatabaseDataError(DatabaseError):
+    pass
+
+
 class SQLiteExtract(Function):
     __slots__ = ()
     _function = 'EXTRACT'
@@ -107,7 +110,7 @@ def date_trunc(_type, date):
             ('hour', 0),
             ('day', 1),
             ('month', 1)]:
-        if _type.startswith(attribute):
+        if _type.lower().startswith(attribute):
             break
         value = value.replace(**{attribute: replace})
     return str(value)
@@ -271,6 +274,14 @@ def least(*args):
         return None
 
 
+def bool_and(*args):
+    return all(args)
+
+
+def bool_or(*args):
+    return any(args)
+
+
 def cbrt(value):
     return math.pow(value, 1 / 3)
 
@@ -329,10 +340,11 @@ class Database(DatabaseInterface):
     IN_MAX = 200
 
     TYPES_MAPPING = {
-        'DATETIME': SQLType('TIMESTAMP', 'TIMESTAMP'),
         'BIGINT': SQLType('INTEGER', 'INTEGER'),
         'BOOL': SQLType('BOOLEAN', 'BOOLEAN'),
+        'DATETIME': SQLType('TIMESTAMP', 'TIMESTAMP'),
         'FULLTEXT': SQLType('TEXT', 'TEXT'),
+        'JSON': SQLType('TEXT', 'TEXT'),
         }
 
     def __new__(cls, name=_default_name):
@@ -362,6 +374,8 @@ class Database(DatabaseInterface):
         self._conn.create_function('now', 0, now)
         self._conn.create_function('greatest', -1, greatest)
         self._conn.create_function('least', -1, least)
+        self._conn.create_function('bool_and', -1, bool_and)
+        self._conn.create_function('bool_or', -1, bool_or)
 
         # Mathematical functions
         self._conn.create_function('cbrt', 1, cbrt)
@@ -463,9 +477,10 @@ class Database(DatabaseInterface):
             cursor = conn.cursor()
             cursor.close()
 
-    def drop(self, connection, database_name):
+    @classmethod
+    def drop(cls, connection, database_name):
         if database_name == ':memory:':
-            self._local.memory_database._conn = None
+            cls._local.memory_database._conn = None
             return
         if os.sep in database_name:
             return
@@ -509,15 +524,12 @@ class Database(DatabaseInterface):
 
             ir_module = Table('ir_module')
             ir_module_dependency = Table('ir_module_dependency')
-            for module in ('ir', 'res'):
-                state = 'not activated'
-                if module in ('ir', 'res'):
-                    state = 'to activate'
+            for module in ['ir', 'res']:
                 info = get_module_info(module)
                 insert = ir_module.insert(
                     [ir_module.create_uid, ir_module.create_date,
                         ir_module.name, ir_module.state],
-                    [[0, CurrentTimestamp(), module, state]])
+                    [[0, CurrentTimestamp(), module, 'to activate']])
                 cursor.execute(*insert)
                 cursor.execute('SELECT last_insert_rowid()')
                 module_id, = cursor.fetchone()
@@ -577,11 +589,14 @@ class Database(DatabaseInterface):
     def has_multirow_insert(self):
         return True
 
+    def has_window_functions(self):
+        return sqlite.sqlite_version_info >= (3, 25, 0)
+
     def sql_type(self, type_):
         if type_ in self.TYPES_MAPPING:
             return self.TYPES_MAPPING[type_]
         if type_.startswith('VARCHAR'):
-            return SQLType('VARCHAR', 'VARCHAR')
+            return SQLType('VARCHAR', type_)
         return SQLType(type_, type_)
 
     def sql_format(self, type_, value):
