@@ -5,28 +5,20 @@ import collections
 import json
 import logging
 import os
-import select
+import selectors
 import threading
 import time
 import uuid
-try:
-    from http import HTTPStatus
-except ImportError:
-    from http import client as HTTPStatus
 from urllib.parse import urljoin
 
-from werkzeug.utils import redirect
-from werkzeug.wrappers import Response
-from werkzeug.exceptions import (
-    NotImplemented as NotImplementedException, BadRequest)
-
 from trytond import backend
-from trytond.wsgi import app
-from trytond.transaction import Transaction
-from trytond.protocols.jsonrpc import JSONEncoder, JSONDecoder
 from trytond.config import config
+from trytond.protocols.jsonrpc import JSONDecoder, JSONEncoder
+from trytond.protocols.wrappers import (
+    HTTPStatus, Response, exceptions, redirect)
 from trytond.tools import resolve
-
+from trytond.transaction import Transaction
+from trytond.wsgi import app
 
 logger = logging.getLogger(__name__)
 
@@ -146,11 +138,12 @@ class LongPollingBus:
     def _listen(cls, database):
         db = backend.Database(database)
         if not db.has_channel():
-            raise NotImplementedException
+            raise exceptions.NotImplemented
 
         logger.info("listening on channel '%s'", cls._channel)
         conn = db.get_connection(autocommit=True)
         pid = os.getpid()
+        selector = selectors.DefaultSelector()
         try:
             cursor = conn.cursor()
             cursor.execute('LISTEN "%s"' % cls._channel)
@@ -158,11 +151,9 @@ class LongPollingBus:
             cls._messages[database] = messages = _MessageQueue(_cache_timeout)
 
             now = time.time()
+            selector.register(conn, selectors.EVENT_READ)
             while cls._queues[pid, database]['timeout'] > now:
-                readable, _, _ = select.select([conn], [], [], _select_timeout)
-                if not readable:
-                    continue
-
+                selector.select(timeout=_select_timeout)
                 conn.poll()
                 while conn.notifies:
                     notification = conn.notifies.pop()
@@ -188,6 +179,7 @@ class LongPollingBus:
                 del cls._queues[pid, database]
             raise
         finally:
+            selector.close()
             db.put_connection(conn)
 
         with cls._queues_lock[pid]:
@@ -226,7 +218,7 @@ else:
 @app.auth_required
 def subscribe(request, database_name):
     if not _allow_subscribe:
-        raise NotImplementedException
+        raise exceptions.NotImplemented
     if _url_host and _url_host != request.host_url:
         response = redirect(
             urljoin(_url_host, request.path), HTTPStatus.PERMANENT_REDIRECT)
@@ -237,7 +229,7 @@ def subscribe(request, database_name):
     user = request.authorization.get('userid')
     channels = request.parsed_data.get('channels', [])
     if user is None:
-        raise BadRequest
+        raise exceptions.BadRequest
 
     channels = set(filter(lambda c: not c.startswith('user:'), channels))
     channels.add('user:%s' % user)
