@@ -4,12 +4,10 @@ import datetime as dt
 import json
 import logging
 import os
-import select
+import selectors
 import threading
-import time
 from collections import OrderedDict, defaultdict
 from copy import deepcopy
-from datetime import datetime
 from weakref import WeakKeyDictionary
 
 from sql import Table
@@ -19,11 +17,12 @@ from sql.functions import CurrentTimestamp, Function
 from trytond import backend
 from trytond.config import config
 from trytond.pool import Pool
+from trytond.tools import grouped_slice, resolve
 from trytond.transaction import Transaction
-from trytond.tools import resolve, grouped_slice
 
 __all__ = ['BaseCache', 'Cache', 'LRUDict', 'LRUDictTransaction']
 _clear_timeout = config.getint('cache', 'clean_timeout', default=5 * 60)
+_default_size_limit = config.getint('cache', 'default')
 logger = logging.getLogger(__name__)
 
 
@@ -47,7 +46,9 @@ def freeze(o):
 
 
 def unfreeze(o):
-    if isinstance(o, frozenset):
+    if isinstance(o, tuple):
+        return [unfreeze(x) for x in o]
+    elif isinstance(o, frozenset):
         return dict((x, unfreeze(y)) for x, y in o)
     else:
         return o
@@ -64,11 +65,24 @@ def _get_modules(cursor):
 
 class BaseCache(object):
     _instances = {}
+    context_ignored_keys = {
+        'client', '_request', '_check_access', '_skip_warnings',
+        }
 
-    def __init__(self, name, size_limit=1024, duration=None, context=True):
+    def __init__(
+            self, name, duration=None, context=True,
+            context_ignored_keys=None):
+        assert ((context_ignored_keys is not None and context)
+            or (context_ignored_keys is None)), (
+                f"context_ignored_keys ({context_ignored_keys}) is not valid"
+                f" in regards to context ({context}).")
         self._name = name
-        self.size_limit = size_limit
+        self.size_limit = config.getint(
+            'cache', name, default=_default_size_limit)
         self.context = context
+        self.context_ignored_keys = set()
+        if context and context_ignored_keys:
+            self.context_ignored_keys.update(context_ignored_keys)
         self.hit = self.miss = 0
         if isinstance(duration, dt.timedelta):
             self.duration = duration
@@ -93,9 +107,10 @@ class BaseCache(object):
     def _key(self, key):
         if self.context:
             context = Transaction().context.copy()
-            context.pop('client', None)
-            context.pop('_request', None)
-            return (key, Transaction().user, freeze(context))
+            for k in (self.__class__.context_ignored_keys
+                    | self.context_ignored_keys):
+                context.pop(k, None)
+            return (key, freeze(context))
         return key
 
     def get(self, key, default=None):
@@ -155,7 +170,7 @@ class MemoryCache(BaseCache):
         transaction = Transaction()
         dbname = transaction.database.name
         lower = self._transaction_lower.get(dbname, self._default_lower)
-        if (transaction in self._reset
+        if (self._name in self._reset.get(transaction, set())
                 or transaction.started_at < lower):
             try:
                 return self._transaction_cache[transaction]
@@ -207,6 +222,11 @@ class MemoryCache(BaseCache):
             Transaction.monotonic_time(),
             self._transaction_lower.get(dbname, self._default_lower))
 
+    @classmethod
+    def _clear_all(cls, dbname):
+        for inst in cls._instances.values():
+            inst._clear(dbname)
+
     @classmethod
     def sync(cls, transaction):
         if cls._clean_last is None:
@@ -222,11 +242,9 @@ class MemoryCache(BaseCache):
                     cls._listener[pid, dbname] = listener = threading.Thread(
                         target=cls._listen, args=(dbname,), daemon=True)
                     listener.start()
-                    while (not getattr(listener, 'listening', False)
-                            and listener.is_alive()):
-                        time.sleep(.01)
             return
-        if (datetime.now() - cls._clean_last).total_seconds() < _clear_timeout:
+        last_clean = (dt.datetime.now() - cls._clean_last).total_seconds()
+        if last_clean < _clear_timeout:
             return
         connection = database.get_connection(readonly=True, autocommit=True)
         try:
@@ -249,7 +267,7 @@ class MemoryCache(BaseCache):
             if not inst_timestamp or timestamp > inst_timestamp:
                 inst._clear(dbname, timestamp)
         Pool(dbname).refresh(modules)
-        cls._clean_last = datetime.now()
+        cls._clean_last = dt.datetime.now()
 
     def sync_since(self, value):
         return self._clean_last > value
@@ -257,7 +275,7 @@ class MemoryCache(BaseCache):
     @classmethod
     def commit(cls, transaction):
         table = Table(cls._table)
-        reset = cls._reset.setdefault(transaction, set())
+        reset = cls._reset.pop(transaction, None)
         if not reset:
             return
         database = transaction.database
@@ -305,15 +323,12 @@ class MemoryCache(BaseCache):
                 connection.commit()
             finally:
                 database.put_connection(connection)
-            cls._clean_last = datetime.now()
+            cls._clean_last = dt.datetime.now()
         reset.clear()
 
     @classmethod
     def rollback(cls, transaction):
-        try:
-            cls._reset[transaction].clear()
-        except KeyError:
-            pass
+        cls._reset.pop(transaction, None)
 
     @classmethod
     def drop(cls, dbname):
@@ -354,7 +369,8 @@ class MemoryCache(BaseCache):
     def _listen(cls, dbname):
         current_thread = threading.current_thread()
         pid = os.getpid()
-        conn = None
+
+        conn, selector = None, None
         try:
             database = backend.Database(dbname)
             if not database.has_channel():
@@ -363,16 +379,18 @@ class MemoryCache(BaseCache):
             logger.info(
                 "listening on channel '%s' of '%s'", cls._channel, dbname)
             conn = database.get_connection(autocommit=True)
+            selector = selectors.DefaultSelector()
 
             cursor = conn.cursor()
             cursor.execute('LISTEN "%s"' % cls._channel)
+            # Clear everything in case we missed a payload
+            Pool(dbname).refresh(_get_modules(cursor))
+            cls._clear_all(dbname)
             current_thread.listening = True
 
+            selector.register(conn, selectors.EVENT_READ)
             while cls._listener.get((pid, dbname)) == current_thread:
-                readable, _, _ = select.select([conn], [], [], 60)
-                if not readable:
-                    continue
-
+                selector.select(timeout=60)
                 conn.poll()
                 while conn.notifies:
                     notification = conn.notifies.pop()
@@ -383,12 +401,14 @@ class MemoryCache(BaseCache):
                         for name in reset:
                             inst = cls._instances[name]
                             inst._clear(dbname)
-                cls._clean_last = datetime.now()
+                cls._clean_last = dt.datetime.now()
         except Exception:
             logger.error(
                 "cache listener on '%s' crashed", dbname, exc_info=True)
             raise
         finally:
+            if selector:
+                selector.close()
             if conn:
                 database.put_connection(conn)
             with cls._listener_lock[pid]:
@@ -408,14 +428,19 @@ class LRUDict(OrderedDict):
     If size limit is reached, it will remove the first added items.
     The default_factory provides the same behavior as in standard
     collections.defaultdict.
+    If default_factory_with_key is set, the default_factory is called with the
+    missing key.
     """
     __slots__ = ('size_limit',)
 
-    def __init__(self, size_limit, default_factory=None, *args, **kwargs):
+    def __init__(self, size_limit,
+            default_factory=None, default_factory_with_key=False,
+            *args, **kwargs):
         assert size_limit > 0
         self.size_limit = size_limit
         super(LRUDict, self).__init__(*args, **kwargs)
         self.default_factory = default_factory
+        self.default_factory_with_key = default_factory_with_key
         self._check_size_limit()
 
     def __setitem__(self, key, value):
@@ -425,7 +450,11 @@ class LRUDict(OrderedDict):
     def __missing__(self, key):
         if self.default_factory is None:
             raise KeyError(key)
-        self[key] = value = self.default_factory()
+        if self.default_factory_with_key:
+            value = self.default_factory(key)
+        else:
+            value = self.default_factory()
+        self[key] = value
         return value
 
     def update(self, *args, **kwargs):
