@@ -1,18 +1,17 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import time
 import datetime
-from xml import sax
 import logging
 import re
-
+import time
 from collections import defaultdict
 from decimal import Decimal
+from xml import sax
 
 from trytond import __version__
-from trytond.pyson import PYSONEncoder, CONTEXT
+from trytond.pyson import CONTEXT, PYSONEncoder
 from trytond.tools import grouped_slice
-from trytond.transaction import Transaction
+from trytond.transaction import Transaction, inactive_records
 
 logger = logging.getLogger(__name__)
 
@@ -20,6 +19,10 @@ CDATA_START = re.compile(r'^\s*\<\!\[cdata\[', re.IGNORECASE)
 CDATA_END = re.compile(r'\]\]\>\s*$', re.IGNORECASE)
 
 
+class ParsingError(Exception):
+    pass
+
+
 class DummyTagHandler:
     """Dubhandler implementing empty methods. Will be used when whe
     want to ignore the xml content"""
@@ -48,7 +51,11 @@ class MenuitemTagHandler:
 
         values = {}
 
-        self.xml_id = attributes['id']
+        try:
+            self.xml_id = attributes['id']
+        except KeyError:
+            self.xml_id = None
+            raise ParsingError("missing 'id' attribute")
 
         for attr in ('name', 'sequence', 'parent', 'action', 'groups'):
             if attr in attributes:
@@ -59,11 +66,18 @@ class MenuitemTagHandler:
             values['active'] = bool(eval(attributes['active']))
 
         if values.get('parent'):
-            values['parent'] = self.mh.get_id(values['parent'])
+            model, id_ = self.mh.get_id(values['parent'])
+            if model != 'ir.ui.menu':
+                raise ParsingError(
+                    "invalid 'ir.ui.menu' parent: %s" % model)
+            values['parent'] = id_
 
         action_name = None
         if values.get('action'):
-            action_id = self.mh.get_id(values['action'])
+            model, action_id = self.mh.get_id(values['action'])
+            if not model.startswith('ir.action'):
+                raise ParsingError(
+                    "invalid model for action: %s" % model)
 
             # TODO maybe use a prefetch for this:
             action = self.mh.pool.get('ir.action').__table__()
@@ -132,12 +146,11 @@ class MenuitemTagHandler:
                 values['icon'] = None
 
         if values.get('groups'):
-            raise Exception("Please use separate records for groups")
+            raise ParsingError("forbidden 'groups' attribute")
 
         if not values.get('name'):
             if not action_name:
-                raise Exception("Please provide at least a 'name' attributes "
-                        "or a 'action' attributes on the menuitem tags.")
+                raise ParsingError("missing 'name' or 'action' attribute")
             else:
                 values['name'] = action_name
 
@@ -158,7 +171,7 @@ class MenuitemTagHandler:
             return None
 
     def current_state(self):
-        return "Tag menuitem with id %s.%s" % (self.mh.module, self.xml_id)
+        return "menuitem '%s.%s'" % (self.mh.module, self.xml_id)
 
 
 class RecordTagHandler:
@@ -182,11 +195,14 @@ class RecordTagHandler:
 
         # Manage the top level tag
         if name == "record":
+            try:
+                self.xml_id = attributes["id"]
+            except KeyError:
+                self.xml_id = None
+                raise ParsingError("missing 'id' attribute")
+
             self.model = self.mh.pool.get(attributes["model"])
-            assert self.model, ("The model %s does not exist !"
-                % (attributes["model"],))
 
-            self.xml_id = attributes["id"]
             self.update = bool(int(attributes.get('update', '0')))
 
             # create/update a dict containing fields values
@@ -224,21 +240,32 @@ class RecordTagHandler:
             context = {}
             context['time'] = time
             context['version'] = __version__.rsplit('.', 1)[0]
-            context['ref'] = self.mh.get_id
+            context['ref'] = lambda xml_id: ','.join(self.mh.get_id(xml_id))
             context['Decimal'] = Decimal
             context['datetime'] = datetime
             if pyson_attr:
                 context.update(CONTEXT)
 
+            field = self.model._fields[field_name]
             if search_attr:
-                search_model = self.model._fields[field_name].model_name
+                search_model = field.model_name
                 SearchModel = self.mh.pool.get(search_model)
-                with Transaction().set_context(active_test=False):
+                with inactive_records():
                     found, = SearchModel.search(eval(search_attr, context))
                     self.values[field_name] = found.id
 
             elif ref_attr:
-                self.values[field_name] = self.mh.get_id(ref_attr)
+                model, id_ = self.mh.get_id(ref_attr)
+                if field._type == 'reference':
+                    self.values[field_name] = '%s,%s' % (model, id_)
+                else:
+                    if (field.model_name == 'ir.action'
+                            and model.startswith('ir.action')):
+                        pass
+                    elif model != field.model_name:
+                        raise ParsingError(
+                            "invalid model for %s: %s" % (field_name, model))
+                    self.values[field_name] = id_
 
             elif eval_attr:
                 value = eval(eval_attr, context)
@@ -247,8 +274,8 @@ class RecordTagHandler:
                 self.values[field_name] = value
 
         else:
-            raise Exception("Tags '%s' not supported inside tag record." %
-                    (name,))
+            raise ParsingError(
+                "forbidden '%s' tag inside record tag" % name)
 
     def characters(self, data):
 
@@ -278,39 +305,21 @@ class RecordTagHandler:
                 self.values[self.current_field] = \
                     CDATA_END.sub('', self.values[self.current_field])
                 self.cdata = 'done'
-
-                value = self.values[self.current_field]
-                match = re.findall(r'[^%]%\((.*?)\)[ds]', value)
-                xml_ids = {}
-                for xml_id in match:
-                    xml_ids[xml_id] = self.mh.get_id(xml_id)
-                self.values[self.current_field] = value % xml_ids
-
             self.current_field = None
             return self
 
         elif name == "record":
             if self.xml_id in self.xml_ids and not self.update:
-                raise Exception('Duplicate id: "%s".' % (self.xml_id,))
+                raise ParsingError("duplicate id: %s" % self.xml_id)
             self.mh.import_record(
                 self.model.__name__, self.values, self.xml_id)
             self.xml_ids.append(self.xml_id)
             return None
         else:
-            raise Exception("Unexpected closing tag '%s'" % (name,))
+            raise ParsingError("unexpected closing tag '%s'" % name)
 
     def current_state(self):
-        return "In tag record model %s with id %s.%s." % \
-               (self.model and self.model.__name__ or "?",
-                   self.mh.module, self.xml_id)
-
-
-# Custom exception:
-class Unhandled_field(Exception):
-    """
-    Raised when a field type is not supported by the update mechanism.
-    """
-    pass
+        return "record '%s.%s'" % (self.mh.module, self.xml_id)
 
 
 class Fs2bdAccessor:
@@ -369,11 +378,8 @@ class Fs2bdAccessor:
         for model in models:
             if model.id in self.browserecord[module][model_name]:
                 for cache in Transaction().cache.values():
-                    for cache in (
-                            cache, cache.get('_language_cache', {}).values()):
-                        if (model_name in cache
-                                and model.id in cache[model_name]):
-                            del cache[model_name][model.id]
+                    if model_name in cache:
+                        cache[model_name].pop(model.id, None)
             self.browserecord[module][model_name][model.id] = model
 
     def fetch_new_module(self, module):
@@ -399,7 +405,7 @@ class Fs2bdAccessor:
                 continue
             self.browserecord[module][model_name] = {}
             for sub_record_ids in grouped_slice(record_ids[model_name]):
-                with Transaction().set_context(active_test=False):
+                with inactive_records():
                     records = Model.search([
                         ('id', 'in', list(sub_record_ids)),
                         ], order=[('id', 'ASC')])
@@ -458,7 +464,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         try:
             self.sax_parser.parse(source)
         except Exception as e:
-            raise Exception("Error " + self.current_state()) from e
+            raise ParsingError("in %s" % self.current_state()) from e
         return self.to_delete
 
     def startElement(self, name, attributes):
@@ -519,7 +525,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         if self.taghandler:
             return self.taghandler.current_state()
         else:
-            return ''
+            return '?'
 
     def get_id(self, xml_id):
 
@@ -529,9 +535,9 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
             module = self.module
 
         if self.fs2db.get(module, xml_id) is None:
-            raise Exception("Reference to %s not found"
-                % ".".join([module, xml_id]))
-        return self.fs2db.get(module, xml_id)["db_id"]
+            raise ParsingError("%s.%s not found" % (module, xml_id))
+        value = self.fs2db.get(module, xml_id)
+        return value['model'], value["db_id"]
 
     @staticmethod
     def _clean_value(key, record):
@@ -552,7 +558,8 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                 return None
             return str(getattr(record, key))
         elif field_type in ['one2many', 'many2many']:
-            raise Unhandled_field("Unhandled field %s" % key)
+            raise ParsingError(
+                "unsupported field %s of type %s" % (key, field_type))
         else:
             return getattr(record, key)
 
@@ -572,7 +579,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         module = self.module
 
         if not fs_id:
-            raise Exception('import_record : Argument fs_id is mandatory')
+            raise ValueError("missing fs_id")
 
         if '.' in fs_id:
             assert len(fs_id.split('.')) == 2, ('"%s" contains too many dots. '
@@ -582,8 +589,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
 
             module, fs_id = fs_id.split('.')
             if not self.fs2db.get(module, fs_id):
-                raise Exception('Reference to %s.%s not found'
-                    % (module, fs_id))
+                raise ParsingError("%s.%s not found" % (module, fs_id))
 
         Model = self.pool.get(model)
 
@@ -598,10 +604,11 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                 return
 
             # this record is already in the db:
-            # XXX maybe use only one call to get()
-            db_id, db_model, mdata_id, old_values = [
-                self.fs2db.get(module, fs_id)[x]
-                for x in ["db_id", "model", "id", "values"]]
+            db_value = self.fs2db.get(module, fs_id)
+            db_id = db_value['db_id']
+            db_model = db_value['model']
+            mdata_id = db_value['id']
+            old_values = db_value['values']
 
             # Check if record has not been deleted
             if db_id is None:
@@ -618,9 +625,8 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                     old_values[key] = old_values[key].decode('utf-8')
 
             if model != db_model:
-                raise Exception("This record try to overwrite "
-                    "data with the wrong model: %s (module: %s)"
-                    % (fs_id, module))
+                raise ParsingError(
+                    "wrong model '%s': %s.%s" % (model, module, fs_id))
 
             record = self.fs2db.get_browserecord(module, Model.__name__, db_id)
             # Re-create record if it was deleted
@@ -787,7 +793,7 @@ def post_import(pool, module, to_delete):
     mdata_delete = []
     ModelData = pool.get("ir.model.data")
 
-    with Transaction().set_context(active_test=False):
+    with inactive_records():
         mdata = ModelData.search([
             ('fs_id', 'in', to_delete),
             ('module', '=', module),
