@@ -1,25 +1,24 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import os
-from operator import itemgetter
 from collections import defaultdict
 from functools import partial
+from operator import itemgetter
 
-from sql import Null
+from genshi.template.text import TextTemplate
 
 from trytond.cache import Cache, MemoryCache
 from trytond.config import config
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelStorage, ModelSQL, DeactivableMixin, fields,
-    sequence_ordered, ModelSingleton)
+    DeactivableMixin, Index, ModelSingleton, ModelSQL, ModelStorage, ModelView,
+    fields, sequence_ordered)
 from trytond.model.exceptions import ValidationError
 from trytond.pool import Pool
-from trytond.pyson import PYSONDecoder, PYSON, Eval
-from trytond.pyson import PYSONEncoder
+from trytond.pyson import PYSON, Eval, PYSONDecoder, PYSONEncoder
 from trytond.rpc import RPC
 from trytond.tools import file_open
-from trytond.transaction import Transaction
+from trytond.transaction import Transaction, inactive_records
 
 if not config.get('html', 'plugins-ir.action.report-report_content_html'):
     config.set(
@@ -47,13 +46,23 @@ class ContextError(ValidationError):
 
 
 EMAIL_REFKEYS = set(('cc', 'to', 'subject'))
+ACTION_SELECTION = [
+    ('ir.action.report', "Report"),
+    ('ir.action.act_window', "Window"),
+    ('ir.action.wizard', "Wizard"),
+    ('ir.action.url', "URL"),
+    ]
 
 
 class Action(DeactivableMixin, ModelSQL, ModelView):
     "Action"
     __name__ = 'ir.action'
     name = fields.Char('Name', required=True, translate=True)
-    type = fields.Char('Type', required=True, readonly=True)
+    type = fields.Selection(
+        ACTION_SELECTION, "Type", required=True, readonly=True)
+    action = fields.Function(
+        fields.Reference("Action", selection=ACTION_SELECTION),
+        'get_action')
     records = fields.Selection([
             ('selected', "Selected"),
             ('listed', "Listed"),
@@ -71,6 +80,9 @@ class Action(DeactivableMixin, ModelSQL, ModelView):
                 'get_action_value': RPC(instantiate=0, cache=dict(days=1)),
                 })
 
+    def get_action(self, name):
+        return f'{self.type},{self.id}'
+
     @classmethod
     def default_records(cls):
         return 'selected'
@@ -86,26 +98,26 @@ class Action(DeactivableMixin, ModelSQL, ModelView):
         pool.get('ir.action.keyword')._get_keyword_cache.clear()
 
     @classmethod
+    @inactive_records
     def get_action_id(cls, action_id):
         pool = Pool()
-        with Transaction().set_context(active_test=False):
-            if cls.search([
-                        ('id', '=', action_id),
-                        ]):
-                return action_id
-            for action_type in (
-                    'ir.action.report',
-                    'ir.action.act_window',
-                    'ir.action.wizard',
-                    'ir.action.url',
-                    ):
-                Action = pool.get(action_type)
-                actions = Action.search([
+        if cls.search([
                     ('id', '=', action_id),
-                    ])
-                if actions:
-                    action, = actions
-                    return action.action.id
+                    ]):
+            return action_id
+        for action_type in (
+                'ir.action.report',
+                'ir.action.act_window',
+                'ir.action.wizard',
+                'ir.action.url',
+                ):
+            Action = pool.get(action_type)
+            actions = Action.search([
+                ('id', '=', action_id),
+                ])
+            if actions:
+                action, = actions
+                return action.action.id
 
     @classmethod
     def get_action_values(cls, type_, action_ids, columns=None):
@@ -152,44 +164,50 @@ class ActionKeyword(ModelSQL, ModelView):
             ], string='Keyword', required=True)
     model = fields.Reference('Model', selection='models_get')
     action = fields.Many2One('ir.action', 'Action',
-        ondelete='CASCADE', select=True)
-    _get_keyword_cache = Cache('ir_action_keyword.get_keyword')
+        ondelete='CASCADE')
+    _get_keyword_cache = Cache(
+        'ir_action_keyword.get_keyword', context=False)
 
     @classmethod
     def __setup__(cls):
         super(ActionKeyword, cls).__setup__()
+        cls.__access__.add('action')
+        table = cls.__table__()
+
         cls.__rpc__.update({
                 'get_keyword': RPC(cache=dict(days=1)),
                 })
+        cls._sql_indexes.add(
+            Index(
+                table,
+                (table.keyword, Index.Equality()),
+                (table.model, Index.Equality())))
 
     @classmethod
-    def __register__(cls, module_name):
-        super(ActionKeyword, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-        table.index_action(['keyword', 'model'], 'add')
+    def validate_fields(cls, actions, field_names):
+        super().validate_fields(actions, field_names)
+        cls.check_wizard_model(actions, field_names)
 
     @classmethod
-    def validate(cls, actions):
-        super(ActionKeyword, cls).validate(actions)
+    def check_wizard_model(cls, actions, field_names=None):
+        pool = Pool()
+        ActionWizard = pool.get('ir.action.wizard')
+        if field_names and not (field_names & {'action', 'model'}):
+            return
         for action in actions:
-            action.check_wizard_model()
-
-    def check_wizard_model(self):
-        ActionWizard = Pool().get('ir.action.wizard')
-        if self.action.type == 'ir.action.wizard':
-            action_wizards = ActionWizard.search([
-                ('action', '=', self.action.id),
-                ], limit=1)
-            # could be empty when copying an action
-            if action_wizards:
-                action_wizard, = action_wizards
-                if action_wizard.model:
-                    if not str(self.model).startswith(
-                            '%s,' % action_wizard.model):
-                        raise WizardModelError(
-                            gettext('ir.msg_action_wrong_wizard_model',
-                                name=action_wizard.rec_name))
+            if action.action.type == 'ir.action.wizard':
+                action_wizards = ActionWizard.search([
+                    ('action', '=', action.action.id),
+                    ], limit=1)
+                # could be empty when copying an action
+                if action_wizards:
+                    action_wizard, = action_wizards
+                    if action_wizard.model:
+                        if not str(action.model).startswith(
+                                '%s,' % action_wizard.model):
+                            raise WizardModelError(
+                                gettext('ir.msg_action_wrong_wizard_model',
+                                    name=action_wizard.rec_name))
 
     @staticmethod
     def _convert_vals(vals):
@@ -208,14 +226,12 @@ class ActionKeyword(ModelSQL, ModelView):
 
     @classmethod
     def delete(cls, keywords):
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         cls._get_keyword_cache.clear()
         super(ActionKeyword, cls).delete(keywords)
 
     @classmethod
     def create(cls, vlist):
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         cls._get_keyword_cache.clear()
         new_vlist = []
@@ -230,7 +246,6 @@ class ActionKeyword(ModelSQL, ModelView):
         for keywords, values in zip(actions, actions):
             args.extend((keywords, cls._convert_vals(values)))
         super(ActionKeyword, cls).write(*args)
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         cls._get_keyword_cache.clear()
 
@@ -239,7 +254,10 @@ class ActionKeyword(ModelSQL, ModelView):
         pool = Pool()
         Action = pool.get('ir.action')
         Menu = pool.get('ir.ui.menu')
-        key = (keyword, tuple(value))
+        ModelAccess = pool.get('ir.model.access')
+        User = pool.get('res.user')
+        groups = User.get_groups()
+        key = (Transaction().language, groups, keyword, tuple(value))
         keywords = cls._get_keyword_cache.get(key)
         if keywords is not None:
             return keywords
@@ -269,6 +287,11 @@ class ActionKeyword(ModelSQL, ModelView):
             types[type_].append(action_keyword.action.id)
         for type_, action_ids in types.items():
             for value in Action.get_action_values(type_, action_ids):
+                if (type_ == 'ir.action.act_window'
+                        and value['res_model']
+                        and not ModelAccess.check(
+                            value['res_model'], raise_exception=False)):
+                    continue
                 value['keyword'] = keyword
                 keywords.append(value)
         if (record_id is not None
@@ -294,10 +317,18 @@ class ActionMixin(ModelSQL):
     _order_name = 'action'
     _action_name = 'name'
 
+    action = fields.Many2One(
+        'ir.action', "Action",
+        required=True, readonly=True, ondelete='CASCADE')
+
     @classmethod
     def __setup__(cls):
         pool = Pool()
         super(ActionMixin, cls).__setup__()
+        cls.__access__.add('action')
+        cls.action.domain = [
+            ('type', '=', cls.__name__),
+            ]
         Action = pool.get('ir.action')
         for name in dir(Action):
             field = getattr(Action, name)
@@ -357,7 +388,6 @@ class ActionMixin(ModelSQL):
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         Action = pool.get('ir.action')
         ir_action = cls.__table__()
@@ -390,8 +420,6 @@ class ActionMixin(ModelSQL):
             cursor.execute(*ir_action.update(
                     [ir_action.id], [action.id],
                     where=ir_action.id == record.id))
-            transaction.database.update_auto_increment(
-                transaction.connection, cls._table, action.id)
             record = cls(action.id)
             new_records.append(record)
             to_write.extend(([record], later))
@@ -404,14 +432,12 @@ class ActionMixin(ModelSQL):
         pool = Pool()
         ActionKeyword = pool.get('ir.action.keyword')
         super(ActionMixin, cls).write(records, values, *args)
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         ActionKeyword._get_keyword_cache.clear()
 
     @classmethod
     def delete(cls, records):
         pool = Pool()
-        ModelView._fields_view_get_cache.clear()
         ModelView._view_toolbar_get_cache.clear()
         Action = pool.get('ir.action')
         actions = [x.action for x in records]
@@ -466,8 +492,6 @@ class ActionReport(ActionMixin, ModelSQL, ModelView):
                 },
             depends=['template_extension']),
         'get_report_content_html', setter='set_report_content_html')
-    action = fields.Many2One('ir.action', 'Action', required=True,
-            ondelete='CASCADE')
     direct_print = fields.Boolean('Direct Print')
     single = fields.Boolean("Single",
         help="Check if the template works only for one record.")
@@ -551,27 +575,13 @@ class ActionReport(ActionMixin, ModelSQL, ModelView):
             ], translate=False,
         string='Extension', help='Leave empty for the same as template, '
         'see LibreOffice documentation for compatible format.')
-    module = fields.Char('Module', readonly=True, select=True)
+    record_name = fields.Char(
+        "Record Name", translate=True,
+        help="A Genshi expression to compute the name using 'record'.\n"
+        "Leave empty for the default name.",)
+    module = fields.Char('Module', readonly=True)
     _template_cache = MemoryCache('ir.action.report.template', context=False)
 
-    @classmethod
-    def __register__(cls, module_name):
-        super(ActionReport, cls).__register__(module_name)
-
-        transaction = Transaction()
-        cursor = transaction.connection.cursor()
-        table = cls.__table_handler__(module_name)
-        action_report = cls.__table__()
-
-        # Migration from 3.4 remove report_name_module_uniq constraint
-        table.drop_constraint('report_name_module_uniq')
-
-        # Migration from 4.4 replace plain extension by txt
-        cursor.execute(*action_report.update(
-                [action_report.extension],
-                ['txt'],
-                where=action_report.extension == 'plain'))
-
     @staticmethod
     def default_type():
         return 'ir.action.report'
@@ -691,6 +701,26 @@ class ActionReport(ActionMixin, ModelSQL, ModelView):
     def set_template_cached(self, template):
         self._template_cache.set(self.id, template)
 
+    @classmethod
+    def validate_fields(cls, reports, field_names):
+        super().validate_fields(reports, field_names)
+        cls.check_record_name(reports, field_names)
+
+    @classmethod
+    def check_record_name(cls, reports, field_names=None):
+        if field_names and 'record_name' not in field_names:
+            return
+        for report in reports:
+            if not report.record_name:
+                return
+            try:
+                TextTemplate(report.record_name)
+            except Exception as exception:
+                raise ValidationError(gettext(
+                        'ir.msg_report_invalid_record_name',
+                        report=report.rec_name,
+                        exception=exception)) from exception
+
 
 class ActionActWindow(ActionMixin, ModelSQL, ModelView):
     "Action act window"
@@ -710,8 +740,6 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
         'act_window', 'Domains')
     domains = fields.Function(fields.Binary('Domains'), 'get_domains')
     limit = fields.Integer('Limit', help='Default limit for the list view.')
-    action = fields.Many2One('ir.action', 'Action', required=True,
-            ondelete='CASCADE')
     search_value = fields.Char('Search Criteria',
             help='Default search criteria for the list view.')
     pyson_domain = fields.Function(fields.Char('PySON Domain'), 'get_pyson')
@@ -728,26 +756,6 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
                 'get': RPC(cache=dict(days=1)),
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        act_window = cls.__table__()
-        super(ActionActWindow, cls).__register__(module_name)
-
-        table = cls.__table_handler__(module_name)
-
-        # Migration from 3.0: auto_refresh removed
-        table.drop_column('auto_refresh')
-
-        # Migration from 4.0: window_name removed
-        table.drop_column('window_name')
-
-        # Migration from 4.2: remove required on limit
-        table.not_null_action('limit', 'remove')
-        cursor.execute(*act_window.update(
-                [act_window.limit], [Null],
-                where=act_window.limit == 0))
-
     @staticmethod
     def default_type():
         return 'ir.action.act_window'
@@ -764,8 +772,12 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
     def validate(cls, actions):
         super(ActionActWindow, cls).validate(actions)
         cls.check_views(actions)
-        cls.check_domain(actions)
-        cls.check_context(actions)
+
+    @classmethod
+    def validate_fields(cls, actions, field_names):
+        super().validate_fields(actions, field_names)
+        cls.check_domain(actions, field_names)
+        cls.check_context(actions, field_names)
 
     @classmethod
     def check_views(cls, actions):
@@ -799,19 +811,21 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
                                 action=action.rec_name))
 
     @classmethod
-    def check_domain(cls, actions):
+    def check_domain(cls, actions, field_names=None):
         "Check domain and search_value"
+        if field_names and not (field_names & {'domain', 'search_value'}):
+            return
         for action in actions:
             for domain in (action.domain, action.search_value):
                 if not domain:
                     continue
                 try:
                     value = PYSONDecoder().decode(domain)
-                except Exception:
+                except Exception as exception:
                     raise DomainError(
                         gettext('ir.msg_action_invalid_domain',
                             domain=domain,
-                            action=action.rec_name))
+                            action=action.rec_name)) from exception
                 if isinstance(value, PYSON):
                     if not value.types() == set([list]):
                         raise DomainError(
@@ -826,24 +840,26 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
                 else:
                     try:
                         fields.domain_validate(value)
-                    except Exception:
+                    except Exception as exception:
                         raise DomainError(
                             gettext('ir.msg_action_invalid_domain',
                                 domain=domain,
-                                action=action.rec_name))
+                                action=action.rec_name)) from exception
 
     @classmethod
-    def check_context(cls, actions):
+    def check_context(cls, actions, field_names=None):
         "Check context"
+        if field_names and 'context' not in field_names:
+            return
         for action in actions:
             if action.context:
                 try:
                     value = PYSONDecoder().decode(action.context)
-                except Exception:
+                except Exception as exception:
                     raise ContextError(
                         gettext('ir.msg_action_invalid_context',
                             context=action.context,
-                            action=action.rec_name))
+                            action=action.rec_name)) from exception
                 if isinstance(value, PYSON):
                     if not value.types() == set([dict]):
                         raise ContextError(
@@ -858,14 +874,14 @@ class ActionActWindow(ActionMixin, ModelSQL, ModelView):
                 else:
                     try:
                         fields.context_validate(value)
-                    except Exception:
+                    except Exception as exception:
                         raise ContextError(
                             gettext('ir.msg_action_invalid_context',
                                 context=action.context,
-                                action=action.rec_name))
+                                action=action.rec_name)) from exception
 
     def get_views(self, name):
-        return [(view.view.id, view.view.type)
+        return [(view.view.id, view.view.rng_type)
             for view in self.act_window_views]
 
     def get_domains(self, name):
@@ -911,10 +927,19 @@ class ActionActWindowView(
         sequence_ordered(), DeactivableMixin, ModelSQL, ModelView):
     "Action act window view"
     __name__ = 'ir.action.act_window.view'
-    view = fields.Many2One('ir.ui.view', 'View', required=True,
-            ondelete='CASCADE')
+    view = fields.Many2One(
+        'ir.ui.view', "View", required=True, ondelete='CASCADE',
+        domain=[
+            ('model', '=', Eval('model')),
+            ])
     act_window = fields.Many2One('ir.action.act_window', 'Action',
             ondelete='CASCADE')
+    model = fields.Function(fields.Char("Model"), 'on_change_with_model')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('act_window')
 
     @classmethod
     def __register__(cls, module_name):
@@ -925,6 +950,11 @@ class ActionActWindowView(
         # Migration from 5.0: remove required on sequence
         table.not_null_action('sequence', 'remove')
 
+    @fields.depends('act_window', '_parent_act_window.res_model')
+    def on_change_with_model(self, name=None):
+        if self.act_window:
+            return self.act_window.res_model
+
     @classmethod
     def create(cls, vlist):
         pool = Pool()
@@ -953,7 +983,12 @@ class ActionActWindowDomain(
     domain = fields.Char('Domain')
     count = fields.Boolean('Count')
     act_window = fields.Many2One('ir.action.act_window', 'Action',
-        select=True, required=True, ondelete='CASCADE')
+        required=True, ondelete='CASCADE')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('act_window')
 
     @classmethod
     def __register__(cls, module_name):
@@ -969,34 +1004,43 @@ class ActionActWindowDomain(
         return False
 
     @classmethod
-    def validate(cls, actions):
-        super(ActionActWindowDomain, cls).validate(actions)
-        cls.check_domain(actions)
+    def validate_fields(cls, actions, field_names):
+        super().validate_fields(actions, field_names)
+        cls.check_domain(actions, field_names)
 
     @classmethod
-    def check_domain(cls, actions):
+    def check_domain(cls, actions, field_names=None):
+        if field_names and 'domain' not in field_names:
+            return
         for action in actions:
             if not action.domain:
                 continue
             try:
                 value = PYSONDecoder().decode(action.domain)
-            except Exception:
-                value = None
+            except Exception as exception:
+                raise DomainError(gettext(
+                        'ir.msg_action_invalid_domain',
+                        domain=action.domain,
+                        action=action.rec_name)) from exception
             if isinstance(value, PYSON):
                 if not value.types() == set([list]):
-                    value = None
+                    raise DomainError(gettext(
+                            'ir.msg_action_invalid_domain',
+                            domain=action.domain,
+                            action=action.rec_name))
             elif not isinstance(value, list):
-                value = None
-            else:
-                try:
-                    fields.domain_validate(value)
-                except Exception:
-                    value = None
-            if value is None:
                 raise DomainError(gettext(
                         'ir.msg_action_invalid_domain',
                         domain=action.domain,
                         action=action.rec_name))
+            else:
+                try:
+                    fields.domain_validate(value)
+                except Exception as exception:
+                    raise DomainError(gettext(
+                            'ir.msg_action_invalid_domain',
+                            domain=action.domain,
+                            action=action.rec_name)) from exception
 
     @classmethod
     def create(cls, vlist):
@@ -1023,8 +1067,6 @@ class ActionWizard(ActionMixin, ModelSQL, ModelView):
     __name__ = 'ir.action.wizard'
     _action_name = 'wiz_name'
     wiz_name = fields.Char('Wizard name', required=True)
-    action = fields.Many2One('ir.action', 'Action', required=True,
-            ondelete='CASCADE')
     model = fields.Char('Model')
     window = fields.Boolean('Window', help='Run wizard in a new window.')
 
@@ -1043,13 +1085,23 @@ class ActionWizard(ActionMixin, ModelSQL, ModelView):
         actions = cls.search(domain)
         return {a.model for a in actions if a.model}
 
+    @classmethod
+    def get_name(cls, name, model):
+        # TODO add cache
+        actions = cls.search([
+                (cls._action_name, '=', name),
+                ('model', '=', model),
+                ], limit=1)
+        if actions:
+            action, = actions
+            return action.name
+        return name
+
 
 class ActionURL(ActionMixin, ModelSQL, ModelView):
     "Action URL"
     __name__ = 'ir.action.url'
     url = fields.Char('Action Url', required=True)
-    action = fields.Many2One('ir.action', 'Action', required=True,
-            ondelete='CASCADE')
 
     @staticmethod
     def default_type():
