@@ -4,28 +4,31 @@ import heapq
 import json
 import logging
 import re
-
-from sql import Null, Literal
-from sql.aggregate import Max
-from sql.conditionals import Case
 from collections import defaultdict
 from itertools import groupby
 
+from sql import Literal, Null
+from sql.aggregate import Max
+from sql.conditionals import Case
+from sql.operators import Equal
+
 from trytond.cache import Cache
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, Workflow, DeactivableMixin, fields, Unique,
-    EvalEnvironment)
+    DeactivableMixin, EvalEnvironment, Exclude, Index, ModelSingleton,
+    ModelSQL, ModelView, Unique, Workflow, fields)
 from trytond.model.exceptions import AccessError, ValidationError
 from trytond.pool import Pool
 from trytond.protocols.jsonrpc import JSONDecoder, JSONEncoder
 from trytond.pyson import Bool, Eval, PYSONDecoder
 from trytond.report import Report
 from trytond.rpc import RPC
-from trytond.tools import is_instance_method, cursor_dict, grouped_slice
+from trytond.tools import cursor_dict, grouped_slice, is_instance_method
 from trytond.tools.string_ import StringMatcher
-from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateView, StateAction, Button
+from trytond.transaction import Transaction, without_check_access
+from trytond.wizard import Button, StateAction, StateView, Wizard
+
+from .resource import ResourceAccessMixin
 
 logger = logging.getLogger(__name__)
 
@@ -72,6 +75,7 @@ class Model(ModelSQL, ModelView):
         cls.__rpc__.update({
                 'list_models': RPC(),
                 'list_history': RPC(),
+                'get_notification': RPC(),
                 'get_names': RPC(),
                 'global_search': RPC(),
                 })
@@ -144,6 +148,15 @@ class Model(ModelSQL, ModelView):
         return [name for name, model in Pool().iterobject()
             if getattr(model, '_history', False)]
 
+    @classmethod
+    def get_notification(cls):
+        "Return a dictionary of model to notify with the depending fields"
+        return {
+            name: list(model._on_change_notify_depends)
+            for name, model in Pool().iterobject()
+            if issubclass(model, ModelView)
+            and model._on_change_notify_depends}
+
     @classmethod
     def get_name_items(cls):
         "Return a list of couple mapping models to names"
@@ -220,7 +233,7 @@ class ModelField(ModelSQL, ModelView):
             },
         depends=['module'])
     model = fields.Many2One('ir.model', 'Model', required=True,
-        select=True, ondelete='CASCADE',
+        ondelete='CASCADE',
         states={
             'readonly': Bool(Eval('module')),
             },
@@ -257,6 +270,7 @@ class ModelField(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ModelField, cls).__setup__()
+        cls.__access__.add('model')
         table = cls.__table__()
         cls._sql_constraints += [
             ('name_model_uniq', Unique(table, table.name, table.model),
@@ -480,13 +494,14 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ModelAccess, cls).__setup__()
+        cls.__access__.add('model')
         cls.__rpc__.update({
                 'get_access': RPC(),
                 })
 
-    @staticmethod
-    def check_xml_record(accesses, values):
-        return True
+    @classmethod
+    def check_xml_record(cls, accesses, values):
+        pass
 
     @staticmethod
     def default_perm_read():
@@ -520,18 +535,16 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
 
         pool = Pool()
         Model = pool.get('ir.model')
-        UserGroup = pool.get('res.user-res.group')
-        Group = pool.get('res.group')
+        User = pool.get('res.user')
         cursor = Transaction().connection.cursor()
-        user = Transaction().user
         model_access = cls.__table__()
         ir_model = Model.__table__()
-        user_group = UserGroup.user_group_all_table()
-        group = Group.__table__()
+
+        groups = User.get_groups()
 
         access = {}
         for model in models:
-            maccess = cls._get_access_cache.get((user, model), default=-1)
+            maccess = cls._get_access_cache.get((groups, model), default=-1)
             if maccess == -1:
                 break
             access[model] = maccess
@@ -551,13 +564,27 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
 
         all_models = list(set(sum(model2models.values(), [])))
         default = {'read': True, 'write': True, 'create': True, 'delete': True}
-        access = dict((m, default) for m in models)
+        default_singleton = {
+            'read': True, 'write': True, 'create': False, 'delete': False}
+        default_table_query = {
+            'read': True, 'write': False, 'create': False, 'delete': False}
+        access = {}
+        for model in models:
+            Model = pool.get(model)
+            if callable(getattr(Model, 'table_query', None)):
+                maccess = access[model] = default_table_query.copy()
+                if Model.create.__func__ != ModelSQL.create.__func__:
+                    maccess['create'] = default['create']
+                if Model.write.__func__ != ModelSQL.write.__func__:
+                    maccess['write'] = default['write']
+                if Model.delete.__func__ != ModelSQL.delete.__func__:
+                    maccess['delete'] = default['delete']
+            elif issubclass(Model, ModelSingleton):
+                access[model] = default_singleton
+            else:
+                access[model] = default
         cursor.execute(*model_access.join(ir_model, 'LEFT',
                 condition=model_access.model == ir_model.id
-                ).join(user_group, 'LEFT',
-                condition=user_group.group == model_access.group
-                ).join(group, 'LEFT',
-                condition=model_access.group == group.id
                 ).select(
                 ir_model.model,
                 Max(Case(
@@ -574,9 +601,7 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
                         else_=0)),
                 where=ir_model.model.in_(all_models)
                 & (model_access.active == Literal(True))
-                & ((
-                        (user_group.user == user)
-                        & (group.active == Literal(True)))
+                & (model_access.group.in_(groups or [-1])
                     | (model_access.group == Null)),
                 group_by=ir_model.model))
         raw_access = {
@@ -588,10 +613,10 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
                 perm: max(
                     (raw_access[m][perm] for m in model2models[model]
                         if m in raw_access),
-                    default=True)
+                    default=access[model][perm])
                 for perm in ['read', 'write', 'create', 'delete']}
         for model, maccess in access.items():
-            cls._get_access_cache.set((user, model), maccess)
+            cls._get_access_cache.set((groups, model), maccess)
         return access
 
     @classmethod
@@ -601,9 +626,9 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
         Model = pool.get(model_name)
         assert mode in ['read', 'write', 'create', 'delete'], \
             'Invalid access mode for security'
-        if ((Transaction().user == 0)
-                or (raise_exception
-                    and not Transaction().context.get('_check_access'))):
+        transaction = Transaction()
+        if (transaction.user == 0
+                or (raise_exception and not transaction.check_access)):
             return True
 
         access = cls.get_access([model_name])[model_name][mode]
@@ -688,9 +713,14 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
     description = fields.Text('Description')
     _get_access_cache = Cache('ir_model_field_access.check', context=False)
 
-    @staticmethod
-    def check_xml_record(field_accesses, values):
-        return True
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('field')
+
+    @classmethod
+    def check_xml_record(cls, field_accesses, values):
+        pass
 
     @staticmethod
     def default_perm_read():
@@ -726,18 +756,16 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
         pool = Pool()
         Model = pool.get('ir.model')
         ModelField = pool.get('ir.model.field')
-        UserGroup = pool.get('res.user-res.group')
-        Group = pool.get('res.group')
-        user = Transaction().user
+        User = pool.get('res.user')
         field_access = cls.__table__()
         ir_model = Model.__table__()
         model_field = ModelField.__table__()
-        user_group = UserGroup.user_group_all_table()
-        group = Group.__table__()
+
+        groups = User.get_groups()
 
         accesses = {}
         for model in models:
-            maccesses = cls._get_access_cache.get((user, model))
+            maccesses = cls._get_access_cache.get((groups, model))
             if maccesses is None:
                 break
             accesses[model] = maccesses
@@ -751,10 +779,6 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
                 condition=field_access.field == model_field.id
                 ).join(ir_model,
                 condition=model_field.model == ir_model.id
-                ).join(user_group, 'LEFT',
-                condition=user_group.group == field_access.group
-                ).join(group, 'LEFT',
-                condition=field_access.group == group.id
                 ).select(
                 ir_model.model,
                 model_field.name,
@@ -772,15 +796,13 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
                         else_=0)),
                 where=ir_model.model.in_(models)
                 & (field_access.active == Literal(True))
-                & ((
-                        (user_group.user == user)
-                        & (group.active == Literal(True)))
+                & (field_access.group.in_(groups or [-1])
                     | (field_access.group == Null)),
                 group_by=[ir_model.model, model_field.name]))
         for m, f, r, w, c, d in cursor:
             accesses[m][f] = {'read': r, 'write': w, 'create': c, 'delete': d}
         for model, maccesses in accesses.items():
-            cls._get_access_cache.set((user, model), maccesses)
+            cls._get_access_cache.set((groups, model), maccesses)
         return accesses
 
     @classmethod
@@ -793,9 +815,9 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
         Model = pool.get(model_name)
         assert mode in ('read', 'write', 'create', 'delete'), \
             'Invalid access mode'
-        if ((Transaction().user == 0)
-                or (raise_exception
-                    and not Transaction().context.get('_check_access'))):
+        transaction = Transaction()
+        if (transaction.user == 0
+                or (raise_exception and not transaction.check_access)):
             if access:
                 return dict((x, True) for x in fields)
             return True
@@ -837,7 +859,7 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
         ModelView._fields_view_get_cache.clear()
 
 
-class ModelButton(ModelSQL, ModelView):
+class ModelButton(DeactivableMixin, ModelSQL, ModelView):
     "Model Button"
     __name__ = 'ir.model.button'
     name = fields.Char('Name', required=True, readonly=True)
@@ -846,10 +868,7 @@ class ModelButton(ModelSQL, ModelView):
     confirm = fields.Text("Confirm", translate=True,
         help="Text to ask user confirmation when clicking the button.")
     model = fields.Many2One('ir.model', 'Model', required=True, readonly=True,
-        ondelete='CASCADE', select=True)
-    groups = fields.Many2Many('ir.model.button-res.group', 'button', 'group',
-        'Groups')
-    _groups_cache = Cache('ir.model.button.groups')
+        ondelete='CASCADE')
     rules = fields.One2Many('ir.model.button.rule', 'button', "Rules")
     _rules_cache = Cache('ir.model.button.rules')
     clicks = fields.One2Many('ir.model.button.click', 'button', "Clicks")
@@ -872,21 +891,31 @@ class ModelButton(ModelSQL, ModelView):
     _view_attributes_cache = Cache(
         'ir.model.button.view_attributes', context=False)
 
+    @classmethod
+    def __register__(cls, module_name):
+        super().__register__(module_name)
+
+        table_h = cls.__table_handler__(module_name)
+
+        # Migration from 6.2: replace unique by exclude
+        table_h.drop_constraint('name_model_uniq')
+
     @classmethod
     def __setup__(cls):
         super(ModelButton, cls).__setup__()
-        table = cls.__table__()
+        cls.__access__.add('model')
+        t = cls.__table__()
         cls._sql_constraints += [
-            ('name_model_uniq', Unique(table, table.name, table.model),
-                'The button name in model must be unique!'),
+            ('name_model_exclude',
+                Exclude(t, (t.name, Equal), (t.model, Equal),
+                    where=(t.active == Literal(True))),
+                'ir.msg_button_name_unique'),
             ]
         cls._order.insert(0, ('model', 'ASC'))
 
     @classmethod
     def create(cls, vlist):
         result = super(ModelButton, cls).create(vlist)
-        # Restart the cache for get_groups and get_rules
-        cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
         cls._view_attributes_cache.clear()
@@ -895,8 +924,6 @@ class ModelButton(ModelSQL, ModelView):
     @classmethod
     def write(cls, buttons, values, *args):
         super(ModelButton, cls).write(buttons, values, *args)
-        # Restart the cache for get_groups and get_rules
-        cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
         cls._view_attributes_cache.clear()
@@ -904,8 +931,6 @@ class ModelButton(ModelSQL, ModelView):
     @classmethod
     def delete(cls, buttons):
         super(ModelButton, cls).delete(buttons)
-        # Restart the cache for get_groups and get_rules
-        cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
         cls._view_attributes_cache.clear()
@@ -920,27 +945,7 @@ class ModelButton(ModelSQL, ModelView):
         return super(ModelButton, cls).copy(buttons, default=default)
 
     @classmethod
-    def get_groups(cls, model, name):
-        '''
-        Return a set of group ids for the named button on the model.
-        '''
-        key = (model, name)
-        groups = cls._groups_cache.get(key)
-        if groups is not None:
-            return groups
-        buttons = cls.search([
-                ('model.model', '=', model),
-                ('name', '=', name),
-                ])
-        if not buttons:
-            groups = set()
-        else:
-            button, = buttons
-            groups = set(g.id for g in button.groups)
-        cls._groups_cache.set(key, groups)
-        return groups
-
-    @classmethod
+    @without_check_access
     def get_rules(cls, model, name):
         'Return a list of rules to apply on the named button of the model'
         pool = Pool()
@@ -962,6 +967,7 @@ class ModelButton(ModelSQL, ModelView):
         return rules
 
     @classmethod
+    @without_check_access
     def get_reset(cls, model, name):
         "Return a list of button names to reset"
         key = (model, name)
@@ -1016,17 +1022,24 @@ class ModelButtonRule(ModelSQL, ModelView):
         help='A PYSON statement evaluated with the record represented by '
         '"self"\nIt activate the rule if true.')
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('button')
+
     @classmethod
     def default_number_user(cls):
         return 1
 
     @classmethod
-    def validate(cls, rules):
-        super(ModelButtonRule, cls).validate(rules)
-        cls.check_condition(rules)
+    def validate_fields(cls, rules, field_names):
+        super().validate_fields(rules, field_names)
+        cls.check_condition(rules, field_names)
 
     @classmethod
-    def check_condition(cls, rules):
+    def check_condition(cls, rules, field_names=None):
+        if field_names and 'condition' not in field_names:
+            return
         for rule in rules:
             if not rule.condition:
                 continue
@@ -1087,6 +1100,7 @@ class ModelButtonClick(DeactivableMixin, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ModelButtonClick, cls).__setup__()
+        cls.__access__.add('button')
         cls.__rpc__.update({
                 'get_click': RPC(),
                 })
@@ -1121,6 +1135,7 @@ class ModelButtonClick(DeactivableMixin, ModelSQL, ModelView):
         return clicks
 
     @classmethod
+    @without_check_access
     def reset(cls, model, names, records):
         assert all(r.__class__.__name__ == model for r in records)
 
@@ -1150,22 +1165,21 @@ class ModelButtonReset(ModelSQL):
     __name__ = 'ir.model.button-button.reset'
     button_ruled = fields.Many2One(
         'ir.model.button', "Button Ruled",
-        required=True, ondelete='CASCADE', select=True)
+        required=True, ondelete='CASCADE')
     button = fields.Many2One(
         'ir.model.button', "Button",
-        required=True, ondelete='CASCADE', select=True)
+        required=True, ondelete='CASCADE')
 
 
 class ModelData(ModelSQL, ModelView):
     "Model data"
     __name__ = 'ir.model.data'
     fs_id = fields.Char('Identifier on File System', required=True,
-        help="The id of the record as known on the file system.",
-        select=True)
-    model = fields.Char('Model', required=True, select=True)
-    module = fields.Char('Module', required=True, select=True)
+        help="The id of the record as known on the file system.")
+    model = fields.Char('Model', required=True)
+    module = fields.Char('Module', required=True)
     db_id = fields.Integer(
-        "Resource ID", select=True,
+        "Resource ID",
         states={
             'required': ~Eval('noupdate', False),
             },
@@ -1186,7 +1200,17 @@ class ModelData(ModelSQL, ModelView):
             ('fs_id_module_model_uniq',
                 Unique(table, table.fs_id, table.module, table.model),
                 'The triple (fs_id, module, model) must be unique!'),
-        ]
+            ]
+        cls._sql_indexes.update({
+                Index(
+                    table,
+                    (table.fs_id, Index.Equality()),
+                    (table.module, Index.Equality()),
+                    (table.model, Index.Equality())),
+                Index(
+                    table,
+                    (table.module, Index.Equality())),
+                })
         cls._buttons.update({
                 'sync': {
                     'invisible': ~Eval('out_of_sync'),
@@ -1200,19 +1224,10 @@ class ModelData(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        model_data = cls.__table__()
-
         super(ModelData, cls).__register__(module_name)
 
         table_h = cls.__table_handler__(module_name)
 
-        # Migration from 4.6: register buttons on ir module
-        cursor.execute(*model_data.update(
-                [model_data.module], ['ir'],
-                where=((model_data.module == 'res')
-                    & (model_data.fs_id == 'model_data_sync_button'))))
-
         # Migration from 5.0: remove required on db_id
         table_h.not_null_action('db_id', action='remove')
 
@@ -1265,6 +1280,55 @@ class ModelData(ModelSQL, ModelView):
         return model in models
 
     @classmethod
+    @without_check_access
+    def can_modify(cls, records, values):
+        for Model, records in groupby(
+                records, key=lambda r: r.__class__):
+            for sub_records in grouped_slice(records):
+                id2record = {r.id: r for r in sub_records}
+                data = cls.search([
+                        ('model', '=', Model.__name__),
+                        ('db_id', 'in', list(id2record.keys())),
+                        ], order=[])
+                for data in data:
+                    record = id2record[data.db_id]
+                    if values is None:
+                        if not data.noupdate:
+                            raise AccessError(
+                                gettext(
+                                    'ir.msg_delete_xml_record',
+                                    **Model.__names__(record=record)),
+                                gettext('ir.msg_base_config_record'))
+                    else:
+                        if not data.values or data.noupdate:
+                            continue
+                        xml_values = cls.load_values(data.values)
+                        for key, val in values.items():
+                            if key in xml_values and val != xml_values[key]:
+                                raise AccessError(
+                                    gettext(
+                                        'ir.msg_write_xml_record',
+                                        **cls.__names__(
+                                            field=key, record=record)),
+                                    gettext('ir.msg_base_config_record'))
+
+    @classmethod
+    @without_check_access
+    def clean(cls, records):
+        data = []
+        for name, records in groupby(
+                records, key=lambda r: r.__class__.__name__):
+            for sub_records in grouped_slice(records):
+                ids = [r.id for r in sub_records]
+                data += cls.search([
+                        ('model', '=', name),
+                        ('db_id', 'in', ids),
+                        ('noupdate', '=', True),
+                        ], order=[])
+        cls.write(data, {'db_id': None})
+
+    @classmethod
+    @without_check_access
     def get_id(cls, module, fs_id=None):
         """
         Return for an fs_id the corresponding db_id.
@@ -1294,16 +1358,7 @@ class ModelData(ModelSQL, ModelView):
 
     @classmethod
     def load_values(cls, values):
-        try:
-            return dict(json.loads(values, object_hook=JSONDecoder()))
-        except ValueError:
-            # Migration from 3.2
-            from decimal import Decimal
-            import datetime
-            return eval(values, {
-                    'Decimal': Decimal,
-                    'datetime': datetime,
-                    })
+        return dict(json.loads(values, object_hook=JSONDecoder()))
 
     @classmethod
     @ModelView.button
@@ -1347,6 +1402,84 @@ class ModelData(ModelSQL, ModelView):
                 cls.write(*to_write)
 
 
+class Log(ResourceAccessMixin, ModelSQL, ModelView):
+    "Log"
+    __name__ = 'ir.model.log'
+
+    user = fields.Many2One(
+        'res.user', "User",
+        states={
+            'required': Eval('event') != 'transition',
+            })
+    event = fields.Selection([
+            ('write', "Modified"),
+            ('delete', "Deleted"),
+            ('button', "Clicked on"),
+            ('wizard', "Launched"),
+            ('transition', "Transitioned to"),
+            ], "Event", required=True)
+    event_string = event.translated('event')
+    target = fields.Char(
+        "Target",
+        states={
+            'required': Eval('event').in_(
+                ['write', 'button', 'wizard', 'transition']),
+            'invisible': (
+                ~Eval('event').in_(
+                    ['write', 'button', 'wizard', 'transition'])),
+            })
+    action = fields.Function(
+        fields.Char("Action"), 'get_action', searcher='search_action')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.resource.required = False  # store deleted record
+        cls._order = [
+            ('create_date', 'DESC'),
+            ('id', 'DESC'),
+            ]
+
+    @classmethod
+    def get_models(cls):
+        return super().get_models() + [(None, '')]  # store deleted record
+
+    def get_action(self, name):
+        pool = Pool()
+        Field = pool.get('ir.model.field')
+        Button = pool.get('ir.model.button')
+        Wizard = pool.get('ir.action.wizard')
+        if self.resource:
+            Model = self.resource.__class__
+            model = self.resource.__name__
+            if self.event == 'write':
+                fields = self.target.split(',')
+                get_name = Field.get_name
+                fields = [get_name(model, f) for f in fields]
+                return ', '.join(fields)
+            elif self.event == 'button':
+                return Button.get_view_attributes(model, self.target).get(
+                    'string', self.target)
+            elif self.event == 'wizard':
+                wiz_name, state_name = self.target.split(':')
+                wiz_name = Wizard.get_name(wiz_name, model)
+                return f'{state_name} @ {wiz_name}'
+            elif self.event == 'transition':
+                field_name, state = self.target.split(':')
+                field = getattr(Model, field_name, None)
+                field_name = Field.get_name(model, field_name)
+                if field:
+                    selection = field.get_selection(
+                        Model, field.name, self.resource)
+                    state = field.get_selection_string(selection, state)
+                return f'{field_name} : {state}'
+        return self.target
+
+    @classmethod
+    def search_action(cls, name, clause):
+        return [('target', *clause[1:])]
+
+
 class PrintModelGraphStart(ModelView):
     'Print Model Graph'
     __name__ = 'ir.model.print_model_graph.start'
