@@ -8,16 +8,16 @@ from sql.operators import NotIn
 from trytond.cache import Cache
 from trytond.exceptions import UserError
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, fields, Unique, sequence_ordered
+from trytond.model import ModelSQL, ModelView, Unique, fields, sequence_ordered
 from trytond.model.exceptions import AccessError
-from trytond.modules import get_module_list, get_module_info
-from trytond.wizard import Wizard, StateView, Button, StateTransition, \
-    StateAction
-from trytond import backend
+from trytond.modules import get_module_info, get_modules
 from trytond.pool import Pool
-from trytond.transaction import Transaction
 from trytond.pyson import Eval
 from trytond.rpc import RPC
+from trytond.tools import grouped_slice
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
 
 class DeactivateDependencyError(UserError):
@@ -92,43 +92,6 @@ class Module(ModelSQL, ModelView):
                     },
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        sql_table = cls.__table__()
-        model_data_sql_table = ModelData.__table__()
-        cursor = Transaction().connection.cursor()
-
-        # Migration from 3.6: remove double module
-        old_table = 'ir_module_module'
-        if backend.TableHandler.table_exist(old_table):
-            backend.TableHandler.table_rename(old_table, cls._table)
-
-        super(Module, cls).__register__(module_name)
-
-        # Migration from 4.0: rename installed to activated
-        cursor.execute(*sql_table.update(
-                [sql_table.state], ['activated'],
-                where=sql_table.state == 'installed'))
-        cursor.execute(*sql_table.update(
-                [sql_table.state], ['not activated'],
-                where=sql_table.state == 'uninstalled'))
-
-        # Migration from 4.6: register buttons on ir module
-        button_fs_ids = [
-            'module_activate_button',
-            'module_activate_cancel_button',
-            'module_deactivate_button',
-            'module_deactivate_cancel_button',
-            'module_upgrade_button',
-            'module_upgrade_cancel_button',
-            ]
-        cursor.execute(*model_data_sql_table.update(
-                [model_data_sql_table.module], ['ir'],
-                where=((model_data_sql_table.module == 'res')
-                    & (model_data_sql_table.fs_id.in_(button_fs_ids)))))
-
     @staticmethod
     def default_state():
         return 'not activated'
@@ -277,61 +240,41 @@ class Module(ModelSQL, ModelView):
 
     @classmethod
     def update_list(cls):
-        'Update the list of available packages'
-        count = 0
-        module_names = get_module_list()
-
-        modules = cls.search([])
-        name2id = dict((m.name, m.id) for m in modules)
-        cls.delete([m for m in modules
-                if m.state != 'activated' and m.name not in module_names])
-
-        # iterate through activated modules and mark them as being so
-        for name in module_names:
-            if name in name2id:
-                module = cls(name2id[name])
-                tryton = get_module_info(name)
-                cls._update_dependencies(module, tryton.get('depends', []))
-                continue
-
-            tryton = get_module_info(name)
-            if not tryton:
-                continue
-            module, = cls.create([{
-                        'name': name,
-                        'state': 'not activated',
-                        }])
-            count += 1
-            cls._update_dependencies(module, tryton.get('depends', []))
-        return count
-
-    @classmethod
-    def _update_dependencies(cls, module, depends=None):
+        "Update the list of available modules"
         pool = Pool()
         Dependency = pool.get('ir.module.dependency')
-        Dependency.delete([x for x in module.dependencies
-            if x.name not in depends])
-        if depends is None:
-            depends = []
-        # Restart Browse Cache for deleted dependencies
-        module = cls(module.id)
-        dependency_names = [x.name for x in module.dependencies]
-        to_create = []
-        for depend in depends:
-            if depend not in dependency_names:
-                to_create.append({
-                        'module': module.id,
-                        'name': depend,
-                        })
-        if to_create:
-            Dependency.create(to_create)
+        module_names = get_modules(with_test=Pool.test)
+        for sub_module_names in grouped_slice(module_names):
+            cls.delete(cls.search([
+                        ('state', '!=', 'activated'),
+                        ('name', 'not in', list(sub_module_names)),
+                        ]))
+        modules = cls.search([])
+        name2module = {m.name: m for m in modules}
+
+        for name in set(module_names) - name2module.keys():
+            name2module[name] = cls(name=name, state=cls.default_state())
+        cls.save(name2module.values())
+
+        to_save, to_delete = [], []
+        for module in name2module.values():
+            depends = set(get_module_info(module.name).get('depends', []))
+            for dependency in module.dependencies:
+                if dependency.name not in depends:
+                    to_delete.append(dependency)
+            for name in depends - {d.name for d in module.dependencies}:
+                to_save.append(Dependency(name=name, module=module))
+        if to_delete:
+            Dependency.delete(to_delete)
+        if to_save:
+            Dependency.save(to_save)
 
 
 class ModuleDependency(ModelSQL, ModelView):
     "Module dependency"
     __name__ = "ir.module.dependency"
     name = fields.Char('Name')
-    module = fields.Many2One('ir.module', 'Module', select=True,
+    module = fields.Many2One('ir.module', 'Module',
        ondelete='CASCADE', required=True)
     state = fields.Function(fields.Selection([
                 ('not activated', 'Not Activated'),
@@ -345,6 +288,7 @@ class ModuleDependency(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ModuleDependency, cls).__setup__()
+        cls.__access__.add('module')
         table = cls.__table__()
         cls._sql_constraints += [
             ('name_module_uniq', Unique(table, table.name, table.module),
@@ -352,24 +296,17 @@ class ModuleDependency(ModelSQL, ModelView):
         ]
 
     @classmethod
-    def __register__(cls, module_name):
-        # Migration from 3.6: remove double module
-        old_table = 'ir_module_module_dependency'
-        if backend.TableHandler.table_exist(old_table):
-            backend.TableHandler.table_rename(old_table, cls._table)
-
-        super(ModuleDependency, cls).__register__(module_name)
-
-    def get_state(self, name):
+    def get_state(cls, dependencies, name):
         pool = Pool()
         Module = pool.get('ir.module')
-        dependencies = Module.search([
-                ('name', '=', self.name),
-                ])
-        if dependencies:
-            return dependencies[0].state
-        else:
-            return 'unknown'
+        modules = []
+        names = [d.name for d in dependencies]
+        for sub_names in grouped_slice(names):
+            modules.extend(Module.search([
+                        ('name', 'in', list(sub_names)),
+                        ]))
+        name2state = {m.name: m.state for m in modules}
+        return {d.id: name2state.get(d.name, 'unknown') for d in dependencies}
 
 
 class ModuleConfigWizardItem(sequence_ordered(), ModelSQL, ModelView):
@@ -380,25 +317,10 @@ class ModuleConfigWizardItem(sequence_ordered(), ModelSQL, ModelView):
     state = fields.Selection([
         ('open', 'Open'),
         ('done', 'Done'),
-        ], string='State', required=True, select=True)
+        ], string="State", required=True, sort=False)
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        model_data = ModelData.__table__()
-
-        # Migration from 3.6: remove double module
-        old_table = 'ir_module_module_config_wizard_item'
-        if backend.TableHandler.table_exist(old_table):
-            backend.TableHandler.table_rename(old_table, cls._table)
-        cursor.execute(*model_data.update(
-                columns=[model_data.model],
-                values=[cls.__name__],
-                where=(model_data.model
-                    == 'ir.module.module.config_wizard.item')))
-
         super(ModuleConfigWizardItem, cls).__register__(module_name)
 
         table = cls.__table_handler__(module_name)
@@ -424,7 +346,7 @@ class ModuleConfigWizardOther(ModelView):
     'Module Config Wizard Other'
     __name__ = 'ir.module.config_wizard.other'
 
-    percentage = fields.Float('Percentage', readonly=True)
+    percentage = fields.Float('Percentage', digits=(1, 2), readonly=True)
 
     @staticmethod
     def default_percentage():
@@ -434,7 +356,7 @@ class ModuleConfigWizardOther(ModelView):
             ('state', '=', 'done'),
             ], count=True)
         all = Item.search([], count=True)
-        return done / all
+        return round(done / all, 2)
 
 
 class ModuleConfigWizardDone(ModelView):
@@ -492,11 +414,21 @@ class ModuleConfigWizard(Wizard):
     def transition_action(self):
         pool = Pool()
         Item = pool.get('ir.module.config_wizard.item')
+        ModelData = pool.get('ir.model.data')
         items = Item.search([
                 ('state', '=', 'open'),
                 ])
         if items:
             return 'other'
+        items = Item.search([
+                ('state', '=', 'done'),
+                ], order=[('write_date', 'DESC')], limit=1)
+        if items:
+            item, = items
+            # module item will re-launch the config wizard
+            # so do not display the done message.
+            if item.id == ModelData.get_id('ir', 'config_wizard_item_module'):
+                return 'end'
         return 'done'
 
     def end(self):
@@ -526,8 +458,9 @@ class ModuleActivateUpgrade(Wizard):
     upgrade = StateTransition()
     done = StateView('ir.module.activate_upgrade.done',
         'ir.module_activate_upgrade_done_view_form', [
-            Button('OK', 'config', 'tryton-ok', default=True),
+            Button("OK", 'next_', 'tryton-ok', default=True),
             ])
+    next_ = StateTransition()
     config = StateAction('ir.act_module_config_wizard')
 
     @classmethod
@@ -573,13 +506,50 @@ class ModuleActivateUpgrade(Wizard):
             Cache.refresh_pool(transaction)
         return 'done'
 
+    def transition_next_(self):
+        pool = Pool()
+        Item = pool.get('ir.module.config_wizard.item')
+        items = Item.search([
+            ('state', '=', 'open'),
+            ], limit=1)
+        if items:
+            return 'config'
+        else:
+            return 'end'
+
+    def end(self):
+        return 'reload menu'
+
 
 class ModuleConfig(Wizard):
     'Configure Modules'
     __name__ = 'ir.module.config'
 
-    start = StateAction('ir.act_module_form')
+    start = StateView('ir.module.config.start',
+        'ir.module_config_start_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Activate", 'activate', 'tryton-ok', default=True),
+            ])
+    activate = StateAction('ir.act_module_activate_upgrade')
 
-    @staticmethod
-    def transition_start():
+    def do_activate(self, action):
+        pool = Pool()
+        Module = pool.get('ir.module')
+        Module.activate(list(self.start.modules))
+        return action, {}
+
+    @classmethod
+    def transition_activate(cls):
         return 'end'
+
+
+class ModuleConfigStart(ModelView):
+    "Configure Modules"
+    __name__ = 'ir.module.config.start'
+
+    modules = fields.Many2Many(
+        'ir.module', None, None, "Modules",
+        domain=[
+            ('name', '!=', 'tests'),
+            ('state', '=', 'not activated'),
+            ])
