@@ -1,23 +1,21 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import os
-import sys
-import importlib
+import configparser
 import itertools
 import logging
-import configparser
-from glob import iglob
+import os
 from collections import defaultdict
-from importlib.machinery import FileFinder, SourceFileLoader, SOURCE_SUFFIXES
+from glob import iglob
 
 from sql import Table
 from sql.functions import CurrentTimestamp
 
+import trytond.convert as convert
 import trytond.tools as tools
 from trytond.config import config
+from trytond.const import MODULES_GROUP
 from trytond.exceptions import MissingDependenciesException
 from trytond.transaction import Transaction
-import trytond.convert as convert
 
 logger = logging.getLogger(__name__)
 
@@ -29,55 +27,6 @@ MODULES_PATH = os.path.abspath(os.path.dirname(__file__))
 
 MODULES = []
 
-EGG_MODULES = {}
-
-
-def update_egg_modules():
-    global EGG_MODULES
-    try:
-        import pkg_resources
-        for ep in pkg_resources.iter_entry_points('trytond.modules'):
-            EGG_MODULES[ep.name] = ep
-    except ImportError:
-        pass
-
-
-update_egg_modules()
-
-
-def import_module(name, fullname=None):
-    if fullname is None:
-        fullname = 'trytond.modules.' + name
-    try:
-        module = importlib.import_module(fullname)
-    except ImportError:
-        if name not in EGG_MODULES:
-            raise
-        ep = EGG_MODULES[name]
-        # Can not use ep.load because modules are declared in an importable
-        # path and it can not import submodule.
-        path = os.path.join(
-            ep.dist.location, *ep.module_name.split('.')[:-1])
-        if not os.path.isdir(path):
-            # Find module in path
-            for path in sys.path:
-                path = os.path.join(
-                    path, *ep.module_name.split('.')[:-1])
-                if os.path.isdir(os.path.join(path, name)):
-                    break
-            else:
-                # When testing modules from setuptools location is the
-                # module directory
-                path = os.path.dirname(ep.dist.location)
-        spec = FileFinder(
-            path, (SourceFileLoader, SOURCE_SUFFIXES)
-            ).find_spec(fullname)
-        if spec.loader:
-            module = spec.loader.load_module()
-        else:
-            raise
-    return module
-
 
 def get_module_info(name):
     "Return the content of the tryton.cfg"
@@ -139,20 +88,19 @@ class Node(list):
         super(Node, self).append(node)
 
 
-def create_graph(module_list):
-    module_list = set(module_list)
+def create_graph(modules):
     all_deps = set()
     graph = Graph()
-    for module in module_list:
+    for module in modules:
         info = get_module_info(module)
         deps = info.get('depends', []) + [
-            d for d in info.get('extras_depend', []) if d in module_list]
+            d for d in info.get('extras_depend', []) if d in modules]
         node = graph.add(module, deps)
         assert node.info is None
         node.info = info
         all_deps.update(deps)
 
-    missing = all_deps - module_list
+    missing = all_deps - modules
     if missing:
         raise MissingDependenciesException(list(missing))
     return graph
@@ -195,6 +143,7 @@ def load_module_graph(graph, pool, update=None, lang=None):
         update = []
     modules_todo = []
     models_to_update_history = set()
+    models_with_indexes = set()
 
     # Load also parent languages
     lang = set(lang)
@@ -239,6 +188,8 @@ def load_module_graph(graph, pool, update=None, lang=None):
                 for model in classes['model']:
                     if hasattr(model, '_history'):
                         models_to_update_history.add(model.__name__)
+                    if hasattr(model, '_update_sql_indexes'):
+                        models_with_indexes.add(model.__name__)
 
                 # Instanciate a new parser for the module
                 tryton_parser = convert.TrytondXmlHandler(
@@ -292,13 +243,21 @@ def load_module_graph(graph, pool, update=None, lang=None):
             ModelField.clean()
             transaction.commit()
 
-        pool.setup_mixin(modules)
+        pool.setup_mixin()
 
-        for model_name in models_to_update_history:
-            model = pool.get(model_name)
-            if model._history:
-                logger.info('history:update %s', model.__name__)
-                model._update_history_table()
+        if update:
+            for model_name in models_with_indexes:
+                model = pool.get(model_name)
+                if model._sql_indexes:
+                    logger.info('index:create %s', model_name)
+                    model._update_sql_indexes()
+            transaction.commit()
+            for model_name in models_to_update_history:
+                model = pool.get(model_name)
+                if model._history:
+                    logger.info('history:update %s', model.__name__)
+                    model._update_history_table()
+            transaction.commit()
 
         # Vacuum :
         while modules_todo:
@@ -312,8 +271,8 @@ def load_module_graph(graph, pool, update=None, lang=None):
     logger.info('all modules loaded')
 
 
-def get_module_list():
-    module_list = set()
+def get_modules(with_test=False):
+    modules = {'ir', 'res'}
     if os.path.exists(MODULES_PATH) and os.path.isdir(MODULES_PATH):
         for file in os.listdir(MODULES_PATH):
             if file.startswith('.'):
@@ -321,16 +280,15 @@ def get_module_list():
             if file == '__pycache__':
                 continue
             if os.path.isdir(OPJ(MODULES_PATH, file)):
-                module_list.add(file)
-    update_egg_modules()
-    module_list.update(EGG_MODULES.keys())
-    module_list.add('ir')
-    module_list.add('res')
-    module_list.add('tests')
-    return list(module_list)
+                modules.add(file)
+    for ep in tools.entry_points().select(group=MODULES_GROUP):
+        modules.add(ep.name)
+    if with_test:
+        modules.add('tests')
+    return modules
 
 
-def register_classes():
+def register_classes(with_test=False):
     '''
     Import modules to register the classes in the Pool
     '''
@@ -338,28 +296,28 @@ def register_classes():
     trytond.ir.register()
     import trytond.res
     trytond.res.register()
-    import trytond.tests
-    trytond.tests.register()
+    if with_test:
+        import trytond.tests
+        trytond.tests.register()
 
-    for node in create_graph(get_module_list()):
-        module = node.name
-        logger.info('%s:registering classes', module)
+    for node in create_graph(get_modules(with_test=with_test)):
+        module_name = node.name
+        logger.info('%s:registering classes', module_name)
 
-        if module in ('ir', 'res', 'tests'):
-            MODULES.append(module)
+        if module_name in ('ir', 'res', 'tests'):
+            MODULES.append(module_name)
             continue
 
-        the_module = import_module(module)
+        module = tools.import_module(module_name)
         # Some modules register nothing in the Pool
-        if hasattr(the_module, 'register'):
-            the_module.register()
-        MODULES.append(module)
+        if hasattr(module, 'register'):
+            module.register()
+        MODULES.append(module_name)
 
 
 def load_modules(
         database_name, pool, update=None, lang=None, activatedeps=False):
     # Do not import backend when importing module
-    from trytond import backend
     res = True
     if update:
         update = update[:]
@@ -372,23 +330,6 @@ def load_modules(
 
         with transaction.set_context(_no_trigger=True), \
                 transaction.connection.cursor() as cursor:
-            # Migration from 3.6: remove double module
-            old_table = 'ir_module_module'
-            new_table = 'ir_module'
-            if backend.TableHandler.table_exist(old_table):
-                backend.TableHandler.table_rename(old_table, new_table)
-
-            # Migration from 4.0: rename installed to activated
-            cursor.execute(*ir_module.select(ir_module.name,
-                    where=ir_module.state.in_(('installed', 'uninstalled'))))
-            if cursor.fetchone():
-                cursor.execute(*ir_module.update(
-                        [ir_module.state], ['activated'],
-                        where=ir_module.state == 'installed'))
-                cursor.execute(*ir_module.update(
-                        [ir_module.state], ['not activated'],
-                        where=ir_module.state == 'uninstalled'))
-
             if update:
                 cursor.execute(*ir_module.select(ir_module.name,
                         where=ir_module.state.in_(('activated', 'to activate',
@@ -397,12 +338,12 @@ def load_modules(
                 cursor.execute(*ir_module.select(ir_module.name,
                         where=ir_module.state.in_(('activated', 'to upgrade',
                                 'to remove'))))
-            module_list = [name for (name,) in cursor]
+            modules = {name for (name,) in cursor}
             graph = None
             while graph is None:
-                module_list += update
+                modules.update(update)
                 try:
-                    graph = create_graph(module_list)
+                    graph = create_graph(modules)
                 except MissingDependenciesException as e:
                     if not activatedeps:
                         raise
