@@ -4,27 +4,24 @@
 import logging
 import pydoc
 import time
-try:
-    from http import HTTPStatus
-except ImportError:
-    from http import client as HTTPStatus
 
-from werkzeug.exceptions import abort
-from werkzeug.wrappers import Response
 from sql import Table
 
-from trytond import security
-from trytond import backend
+from trytond import __version__, backend, security
 from trytond.config import config, get_hostname
-from trytond import __version__
-from trytond.transaction import Transaction
 from trytond.exceptions import (
-    UserError, UserWarning, ConcurrencyException, LoginException,
-    RateLimitException)
+    ConcurrencyException, LoginException, RateLimitException, UserError,
+    UserWarning)
+from trytond.rpc import RPCReturnException
 from trytond.tools import is_instance_method
-from trytond.wsgi import app
+from trytond.tools.logging import format_args
+from trytond.transaction import Transaction, TransactionError
 from trytond.worker import run_task
-from .wrappers import with_pool
+from trytond.wsgi import app
+
+from .wrappers import HTTPStatus, Response, abort, with_pool
+
+__all__ = ['register_authentication_service']
 
 logger = logging.getLogger(__name__)
 
@@ -48,11 +45,6 @@ def rpc(request, database_name):
 
 
 def login(request, database_name, user, parameters, language=None):
-    try:
-        backend.Database(database_name).connect()
-    except backend.DatabaseOperationalError:
-        logger.error('fail to connect to %s', database_name, exc_info=True)
-        abort(HTTPStatus.NOT_FOUND)
     context = {
         'language': language,
         '_request': request.context,
@@ -61,6 +53,9 @@ def login(request, database_name, user, parameters, language=None):
         session = security.login(
             database_name, user, parameters, context=context)
         code = HTTPStatus.UNAUTHORIZED
+    except backend.DatabaseOperationalError:
+        logger.error('fail to connect to %s', database_name, exc_info=True)
+        abort(HTTPStatus.NOT_FOUND)
     except RateLimitException:
         session = None
         code = HTTPStatus.TOO_MANY_REQUESTS
@@ -82,12 +77,14 @@ def root(request, *args):
     methods = {
         'common.server.version': lambda *a: __version__,
         'common.db.list': db_list,
+        'common.authentication.services': authentication_services,
         }
     return methods[request.rpc_method](request, *request.rpc_params)
 
 
+@app.route('/', methods=['OPTIONS'])
 @app.route('/<path:path>', methods=['OPTIONS'])
-def options(request, path):
+def options(request, path=None):
     return Response(status=HTTPStatus.NO_CONTENT)
 
 
@@ -110,6 +107,17 @@ def db_list(request, *args):
         return transaction.database.list(hostname=hostname)
 
 
+def authentication_services(request):
+    return _AUTHENTICATION_SERVICES
+
+
+def register_authentication_service(name, url):
+    _AUTHENTICATION_SERVICES.append((name, url))
+
+
+_AUTHENTICATION_SERVICES = []
+
+
 @app.auth_required
 @with_pool
 def list_method(request, pool):
@@ -156,25 +164,33 @@ def _dispatch(request, pool, *args, **kwargs):
                 pool.database_name, user, session, context=context):
             abort(HTTPStatus.UNAUTHORIZED)
 
-    log_message = '%s.%s(*%s, **%s) from %s@%s%s'
+    log_message = '%s.%s%s from %s@%s%s in %i ms'
     username = request.authorization.username
     if isinstance(username, bytes):
         username = username.decode('utf-8')
     log_args = (
-        obj, method, args, kwargs, username, request.remote_addr, request.path)
-    logger.debug(log_message, *log_args)
+        obj.__name__, method,
+        format_args(args, kwargs, logger.isEnabledFor(logging.DEBUG)),
+        username, request.remote_addr, request.path)
+
+    def duration():
+        return (time.monotonic() - started) * 1000
+    started = time.monotonic()
 
     retry = config.getint('database', 'retry')
-    for count in range(retry, -1, -1):
-        if count != retry:
+    count = 0
+    transaction_extras = {}
+    while True:
+        if count:
             time.sleep(0.02 * (retry - count))
-        with Transaction().start(pool.database_name, user,
-                readonly=rpc.readonly) as transaction:
+        with Transaction().start(
+                pool.database_name, user, readonly=rpc.readonly,
+                **transaction_extras) as transaction:
             try:
                 c_args, c_kwargs, transaction.context, transaction.timestamp \
                     = rpc.convert(obj, *args, **kwargs)
                 transaction.context['_request'] = request.context
-                meth = getattr(obj, method)
+                meth = rpc.decorate(getattr(obj, method))
                 if (rpc.instantiate is None
                         or not is_instance_method(obj, method)):
                     result = rpc.result(meth(*c_args, **c_kwargs))
@@ -186,19 +202,32 @@ def _dispatch(request, pool, *args, **kwargs):
                     else:
                         result = [rpc.result(meth(i, *c_args, **c_kwargs))
                             for i in inst]
+            except TransactionError as e:
+                transaction.rollback()
+                transaction.tasks.clear()
+                e.fix(transaction_extras)
+                continue
             except backend.DatabaseOperationalError:
-                if count and not rpc.readonly:
+                if count < retry and not rpc.readonly:
                     transaction.rollback()
                     transaction.tasks.clear()
+                    count += 1
+                    logger.debug("Retry: %i", count)
                     continue
-                logger.error(log_message, *log_args, exc_info=True)
+                logger.exception(log_message, *log_args, duration())
                 raise
+            except RPCReturnException as e:
+                transaction.rollback()
+                transaction.tasks.clear()
+                result = e.result()
             except (ConcurrencyException, UserError, UserWarning,
                     LoginException):
-                logger.debug(log_message, *log_args, exc_info=True)
+                logger.info(
+                    log_message, *log_args, duration(),
+                    exc_info=logger.isEnabledFor(logging.DEBUG))
                 raise
             except Exception:
-                logger.error(log_message, *log_args, exc_info=True)
+                logger.exception(log_message, *log_args, duration())
                 raise
             # Need to commit to unlock SQLite database
             transaction.commit()
@@ -208,7 +237,8 @@ def _dispatch(request, pool, *args, **kwargs):
         if session:
             context = {'_request': request.context}
             security.reset(pool.database_name, session, context=context)
-        logger.debug('Result: %s', result)
+        logger.info(log_message, *log_args, duration())
+        logger.debug('Result: %r', result)
         response = app.make_response(request, result)
         if rpc.readonly and rpc.cache:
             response.headers.extend(rpc.cache.headers())
