@@ -4,23 +4,41 @@ import base64
 import gzip
 import logging
 import time
-from io import BytesIO
 from functools import wraps
+from io import BytesIO
+
 try:
     from http import HTTPStatus
 except ImportError:
     from http import client as HTTPStatus
 
-from werkzeug.wrappers import Request as _Request, Response
+from werkzeug import exceptions
 from werkzeug.datastructures import Authorization
-from werkzeug.exceptions import abort, HTTPException
+from werkzeug.exceptions import abort
+from werkzeug.utils import redirect
+from werkzeug.wrappers import Request as _Request
+from werkzeug.wrappers import Response
 
-from trytond import security, backend
-from trytond.exceptions import RateLimitException
+from trytond import backend, security
+from trytond.config import config
+from trytond.exceptions import RateLimitException, UserError, UserWarning
 from trytond.pool import Pool
 from trytond.tools import cached_property
-from trytond.transaction import Transaction
-from trytond.config import config
+from trytond.transaction import Transaction, TransactionError, check_access
+
+__all__ = [
+    'HTTPStatus',
+    'Request',
+    'Response',
+    'abort',
+    'allow_null_origin',
+    'exceptions',
+    'redirect',
+    'set_max_request_size',
+    'user_application',
+    'with_pool',
+    'with_transaction',
+    ]
 
 logger = logging.getLogger(__name__)
 
@@ -178,7 +196,23 @@ def with_pool(func):
         if database_name not in database_list:
             with Transaction().start(database_name, 0, readonly=True):
                 pool.init()
-        return func(request, pool, *args, **kwargs)
+        try:
+            return func(request, pool, *args, **kwargs)
+        except exceptions.HTTPException:
+            logger.debug('%s', request, exc_info=True)
+            raise
+        except (UserError, UserWarning) as e:
+            logger.debug('%s', request, exc_info=True)
+            if request.rpc_method:
+                raise
+            else:
+                abort(HTTPStatus.BAD_REQUEST, e)
+        except Exception as e:
+            logger.error('%s', request, exc_info=True)
+            if request.rpc_method:
+                raise
+            else:
+                abort(HTTPStatus.INTERNAL_SERVER_ERROR, e)
     return wrapper
 
 
@@ -204,25 +238,27 @@ def with_transaction(readonly=None, user=0, context=None):
             else:
                 user_ = user
             retry = config.getint('database', 'retry')
-            for count in range(retry, -1, -1):
-                if count != retry:
+            count = 0
+            transaction_extras = {}
+            while True:
+                if count:
                     time.sleep(0.02 * (retry - count))
                 with Transaction().start(
                         pool.database_name, user_, readonly=readonly_,
-                        context=context_) as transaction:
+                        context=context_, **transaction_extras) as transaction:
                     try:
                         result = func(request, pool, *args, **kwargs)
+                    except TransactionError as e:
+                        transaction.rollback()
+                        transaction.tasks.clear()
+                        e.fix(transaction_extras)
+                        continue
                     except backend.DatabaseOperationalError:
-                        if count and not readonly_:
+                        if count < retry and not readonly_:
                             transaction.rollback()
                             transaction.tasks.clear()
+                            count += 1
                             continue
-                        logger.error('%s', request, exc_info=True)
-                        raise
-                    except Exception as e:
-                        if isinstance(e, HTTPException):
-                            raise
-                        logger.error('%s', request, exc_info=True)
                         raise
                     # Need to commit to unlock SQLite database
                     transaction.commit()
@@ -235,7 +271,8 @@ def with_transaction(readonly=None, user=0, context=None):
 
 
 def user_application(name, json=True):
-    from .jsonrpc import JSONEncoder, json as json_
+    from .jsonrpc import JSONEncoder
+    from .jsonrpc import json as json_
 
     def decorator(func):
         @wraps(func)
@@ -259,14 +296,8 @@ def user_application(name, json=True):
             transaction = Transaction()
             # TODO language
             with transaction.set_user(application.user.id), \
-                    transaction.set_context(_check_access=True):
-                try:
-                    response = func(request, *args, **kwargs)
-                except Exception as e:
-                    if isinstance(e, HTTPException):
-                        raise
-                    logger.error('%s', request, exc_info=True)
-                    abort(HTTPStatus.INTERNAL_SERVER_ERROR, e)
+                    check_access():
+                response = func(request, *args, **kwargs)
             if not isinstance(response, Response) and json:
                 response = Response(json_.dumps(response, cls=JSONEncoder),
                     content_type='application/json')
