@@ -1,9 +1,11 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from trytond.model import ModelSQL, ModelStorage as ModelStorage_, fields
+from trytond.model import ModelSQL
+from trytond.model import ModelStorage as ModelStorage_
+from trytond.model import fields
+from trytond.pool import Pool
 from trytond.pyson import Eval, If
 from trytond.transaction import Transaction
-from trytond.pool import Pool
 
 
 class ModelStorage(ModelSQL):
@@ -18,6 +20,28 @@ class ModelStorageRequired(ModelSQL):
     name = fields.Char('Name', required=True)
 
 
+class ModelStorageSaveMany2One(ModelSQL):
+    "ModelStorage to test save with Many2One"
+    __name__ = 'test.modelstorage.save_m2o'
+    target = fields.Many2One('test.modelstorage.save_m2o.target', "Target")
+    targets = fields.One2Many(
+        'test.modelstorage.save_m2o.target', 'parent', "Targets")
+
+
+class ModelStorageSaveMany2OneTarget(ModelSQL):
+    "ModelStorage to test save with Many2One - Target"
+    __name__ = 'test.modelstorage.save_m2o.target'
+
+    parent = fields.Many2One('test.modelstorage.save_m2o', "Parent")
+    target = fields.Many2One(
+        'test.modelstorage.save_m2o.o2m_target.target', "Target")
+
+
+class ModelStorageSaveMany2OneTargetTarget(ModelSQL):
+    "ModelStorage to test save with Many2One - Target of Target"
+    __name__ = 'test.modelstorage.save_m2o.o2m_target.target'
+
+
 class ModelStorageSave2Many(ModelSQL):
     "Model Storage to test save with xxx2many"
     __name__ = 'test.modelstorage.save2many'
@@ -78,12 +102,43 @@ class ModelStorageRelationDomain(ModelSQL):
                 ('value', '!=', 'valid')),
             ],
         depends=['relation_valid'])
+    relation_context = fields.Many2One(
+        'test.modelstorage.relation_domain.target', "Value",
+        domain=[
+            ('valid', '=', True),
+            ],
+        context={
+            'valid': Eval('relation_valid', True),
+            })
+    relation_pyson_context = fields.Many2One(
+        'test.modelstorage.relation_domain.target', "Value",
+        domain=[
+            If(Eval('relation_valid', True),
+                ('valid', '=', True),
+                ('valid', '=', None)),
+            ],
+        context={
+            'valid': Eval('relation_valid', True),
+            },
+        depends=['relation_valid'])
 
 
 class ModelStorageRelationDomainTarget(ModelSQL):
     "Target of Model stored containing a relation field with a domain"
     __name__ = 'test.modelstorage.relation_domain.target'
     value = fields.Char("Value")
+    valid = fields.Function(
+        fields.Boolean("Valid"), 'get_valid', searcher='search_valid')
+
+    def get_valid(self, name):
+        return Transaction().context.get('valid')
+
+    @classmethod
+    def search_valid(cls, name, domain):
+        if Transaction().context.get('valid') == domain[2]:
+            return []
+        else:
+            return [('id', '=', -1)]
 
 
 class ModelStorageRelationMultiDomain(ModelSQL):
@@ -139,10 +194,21 @@ class ModelStorageEvalEnvironment(ModelStorage_):
         'test.modelstorage.eval_environment', 'many2one', "One2Many")
 
 
+class ModelStorageDomainNotRequired(ModelSQL):
+    "Model for domain on not required field"
+    __name__ = 'test.modelstorage.domain_not_required'
+
+    domain_not_required = fields.Integer(
+        "Domain Not Required", domain=[('domain_not_required', '>', 0)])
+
+
 def register(module):
     Pool.register(
         ModelStorage,
         ModelStorageRequired,
+        ModelStorageSaveMany2One,
+        ModelStorageSaveMany2OneTarget,
+        ModelStorageSaveMany2OneTargetTarget,
         ModelStorageSave2Many,
         ModelStorageSave2ManyTarget,
         ModelStorageSave2ManyRelation,
@@ -155,4 +221,5 @@ def register(module):
         ModelStorageRelationDomain2,
         ModelStorageRelationDomain2Target,
         ModelStorageEvalEnvironment,
+        ModelStorageDomainNotRequired,
         module=module, type_='model')
