@@ -4,9 +4,9 @@
 import unittest
 
 from trytond.model.exceptions import AccessError
+from trytond.pool import Pool
 from trytond.tests.test_tryton import activate_module, with_transaction
 from trytond.transaction import Transaction
-from trytond.pool import Pool
 
 _context = {'_check_access': True}
 
@@ -1030,20 +1030,66 @@ class ModelFieldAccessWriteTestCase(_ModelFieldAccessTestCase):
             TestAccess.write([record], {'field2': 'test'})
 
 
-def suite():
-    suite_ = unittest.TestSuite()
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelAccessReadTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelAccessWriteTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelAccessCreateTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelAccessDeleteTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelAccessModelTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelFieldAccessReadTestCase))
-    suite_.addTests(unittest.TestLoader(
-        ).loadTestsFromTestCase(ModelFieldAccessWriteTestCase))
-    return suite_
+class MenuActionAccessReadTestCase(unittest.TestCase):
+
+    @classmethod
+    def setUpClass(cls):
+        activate_module('tests')
+
+    def create_menu(self, access=None):
+        pool = Pool()
+        Action = pool.get('ir.action.act_window')
+        Menu = pool.get('ir.ui.menu')
+        ModelAccess = pool.get('ir.model.access')
+
+        action = Action(name="Test", res_model='test.access')
+        action.save()
+        menu = Menu(name="Test", action=action)
+        menu.save()
+
+        if access is not None:
+            ModelAccess.create([{
+                        'model': self.model.id,
+                        'perm_read': access,
+                        }])
+        return menu
+
+    @property
+    def model(self):
+        pool = Pool()
+        Model = pool.get('ir.model')
+        model, = Model.search([('model', '=', 'test.access')])
+        return model
+
+    @with_transaction(context=_context)
+    def test_access_empty(self):
+        "Search menu without model access"
+        pool = Pool()
+        Menu = pool.get('ir.ui.menu')
+
+        menu = self.create_menu()
+
+        self.assertEqual(Menu.search([('id', '=', menu.id)]), [menu])
+
+    @with_transaction(context=_context)
+    def test_access(self):
+        "Search menu with model access"
+        pool = Pool()
+        Menu = pool.get('ir.ui.menu')
+
+        menu = self.create_menu(True)
+
+        self.assertEqual(Menu.search([('id', '=', menu.id)]), [menu])
+
+    @with_transaction(context=_context)
+    def test_no_access(self):
+        "Search menu with no model access"
+        pool = Pool()
+        Menu = pool.get('ir.ui.menu')
+
+        menu = self.create_menu(False)
+
+        self.assertEqual(Menu.search([('id', '=', menu.id)]), [])
+
+
+del _ModelAccessTestCase, _ModelFieldAccessTestCase
