@@ -7,8 +7,9 @@ import unittest
 from sql import Literal
 
 from trytond.config import config
-from trytond.model.exceptions import RequiredValidationError
 from trytond.model import fields
+from trytond.model.exceptions import (
+    RequiredValidationError, SQLConstraintError)
 from trytond.pool import Pool
 from trytond.tests.test_tryton import activate_module, with_transaction
 from trytond.transaction import Transaction
@@ -88,6 +89,22 @@ class FieldBinaryTestCase(unittest.TestCase):
                         'binary': cast(b''),
                         }])
 
+    @with_transaction()
+    def test_create_required_with_invalid_sql_constraint(self):
+        "Test create required binary with invalid SQL constraint"
+        Binary = Pool().get('test.binary_required_sql_constraint')
+
+        binary, = Binary.create([{
+                    'binary': cast(b'foo'),
+                    'constraint': True,
+                    }])
+
+        with self.assertRaises(SQLConstraintError):
+            binary, = Binary.create([{
+                        'binary': cast(b'foo'),
+                        'constraint': False,
+                        }])
+
     @with_transaction()
     def test_create_filestorage(self):
         "Test create binary with filestorage"
@@ -123,6 +140,16 @@ class FieldBinaryTestCase(unittest.TestCase):
 
         self.assertEqual(binary.binary, cast(b'foo'))
 
+    @with_transaction()
+    def test_set_sql_value(self):
+        "Test cannot set SQL value"
+        Binary = Pool().get('test.binary')
+
+        binary = Binary()
+
+        with self.assertRaises(ValueError):
+            binary.binary = Literal('foo')
+
     @with_transaction()
     def test_read_size(self):
         "Test read binary size"
@@ -193,6 +220,50 @@ class FieldBinaryTestCase(unittest.TestCase):
         self.assertEqual(binary.binary, None)
         self.assertEqual(binary.binary_id, None)
 
+    @with_transaction()
+    def test_copy(self):
+        "Test copy binary"
+        pool = Pool()
+        Binary = pool.get('test.binary')
+        binary = Binary(binary=b'foo')
+        binary.save()
 
-def suite():
-    return unittest.TestLoader().loadTestsFromTestCase(FieldBinaryTestCase)
+        copy, = Binary.copy([binary])
+
+        self.assertEqual(binary.binary, copy.binary)
+
+    @with_transaction()
+    def test_copy_with_default(self):
+        "Test copy binary with default"
+        pool = Pool()
+        Binary = pool.get('test.binary')
+        binary = Binary(binary=b'foo')
+        binary.save()
+
+        copy, = Binary.copy([binary], default={'binary': b'bar'})
+
+        self.assertEqual(copy.binary, b'bar')
+
+    @with_transaction()
+    def test_copy_with_filestorage(self):
+        "Test copy binary with filestorage"
+        pool = Pool()
+        Binary = pool.get('test.binary_filestorage')
+        binary = Binary(binary=b'foo')
+        binary.save()
+
+        copy, = Binary.copy([binary])
+
+        self.assertEqual(binary.binary, copy.binary)
+
+    @with_transaction()
+    def test_copy_with_filestorage_default(self):
+        "Test copy binary with filestorage and default"
+        pool = Pool()
+        Binary = pool.get('test.binary_filestorage')
+        binary = Binary(binary=b'foo')
+        binary.save()
+
+        copy, = Binary.copy([binary], default={'binary': b'bar'})
+
+        self.assertEqual(copy.binary, b'bar')
