@@ -8,24 +8,50 @@ import importlib
 import io
 import os
 import re
-import sys
 import types
 import unicodedata
 import warnings
 from array import array
+from collections.abc import Sized
 from functools import wraps
-from itertools import islice
+from itertools import islice, tee, zip_longest
 
 from sql import Literal
 from sql.conditionals import Case
 from sql.operators import Or
 
-from trytond.const import OPERATORS
+from trytond.const import MODULES_GROUP, OPERATORS
+
+try:
+    from backports.entry_points_selectable import entry_points as _entry_points
+except ImportError:
+    from importlib.metadata import entry_points as _entry_points
+_ENTRY_POINTS = None
+
+
+def entry_points():
+    global _ENTRY_POINTS
+    if _ENTRY_POINTS is None:
+        _ENTRY_POINTS = _entry_points()
+    return _ENTRY_POINTS
+
+
+def import_module(name):
+    try:
+        ep, = entry_points().select(group=MODULES_GROUP, name=name)
+    except ValueError:
+        return importlib.import_module(f'{MODULES_GROUP}.{name}')
+    return ep.load()
 
 
 def file_open(name, mode="r", subdir='modules', encoding=None):
-    """Open a file from the root dir, using a subdir folder."""
-    from trytond.modules import EGG_MODULES
+    "Open a file from the root directory, using subdir folder"
+    path = find_path(name, subdir, _test=None)
+    return io.open(path, mode, encoding=encoding)
+
+
+def find_path(name, subdir='modules', _test=os.path.isfile):
+    "Return path from the root directory, using subdir folder"
     root_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
 
     def secure_join(root, *paths):
@@ -36,46 +62,37 @@ def file_open(name, mode="r", subdir='modules', encoding=None):
             raise IOError("Permission denied: %s" % name)
         return path
 
-    egg_name = False
-    if subdir == 'modules':
-        module_name = name.split(os.sep)[0]
-        if module_name in EGG_MODULES:
-            epoint = EGG_MODULES[module_name]
-            mod_path = os.path.join(epoint.dist.location,
-                    *epoint.module_name.split('.')[:-1])
-            mod_path = os.path.abspath(mod_path)
-            egg_name = secure_join(mod_path, name)
-            if not os.path.isfile(egg_name):
-                # Find module in path
-                for path in sys.path:
-                    mod_path = os.path.join(path,
-                            *epoint.module_name.split('.')[:-1])
-                    mod_path = os.path.abspath(mod_path)
-                    egg_name = secure_join(mod_path, name)
-                    if os.path.isfile(egg_name):
-                        break
-                if not os.path.isfile(egg_name):
-                    # When testing modules from setuptools location is the
-                    # module directory
-                    egg_name = secure_join(
-                        os.path.dirname(epoint.dist.location), name)
-
     if subdir:
-        if (subdir == 'modules'
-                and (name.startswith('ir' + os.sep)
-                    or name.startswith('res' + os.sep)
-                    or name.startswith('tests' + os.sep))):
-            name = secure_join(root_path, name)
+        if subdir == 'modules':
+            try:
+                module_name, module_path = name.split(os.sep, 1)
+            except ValueError:
+                module_name, module_path = name, ''
+            if module_name in {'ir', 'res', 'tests'}:
+                path = secure_join(root_path, module_name, module_path)
+            else:
+                try:
+                    module = import_module(module_name)
+                except ModuleNotFoundError:
+                    path = secure_join(root_path, subdir, name)
+                else:
+                    path = os.path.dirname(module.__file__)
+                    if module_path:
+                        path = secure_join(path, module_path)
         else:
-            name = secure_join(root_path, subdir, name)
+            path = secure_join(root_path, subdir, name)
     else:
-        name = secure_join(root_path, name)
+        path = secure_join(root_path, name)
 
-    for i in (name, egg_name):
-        if i and os.path.isfile(i):
-            return io.open(i, mode, encoding=encoding)
+    if not _test or _test(path):
+        return path
+    else:
+        raise FileNotFoundError("No such file or directory: %r" % name)
 
-    raise IOError('File not found : %s ' % name)
+
+def find_dir(name, subdir='modules'):
+    "Return directory from the root directory, using subdir folder"
+    return find_path(name, subdir=subdir, _test=os.path.isdir)
 
 
 def get_smtp_server():
@@ -96,12 +113,16 @@ def reduce_ids(field, ids):
     '''
     Return a small SQL expression for the list of ids and the sql column
     '''
-    ids = list(ids)
+    if __debug__:
+        def strict_int(value):
+            assert not isinstance(value, float) or value.is_integer(), \
+                "ids must be integer"
+            return int(value)
+    else:
+        strict_int = int
+    ids = list(map(strict_int, ids))
     if not ids:
         return Literal(False)
-    assert all(x.is_integer() for x in ids if isinstance(x, float)), \
-        'ids must be integer'
-    ids = list(map(int, ids))
     ids.sort()
     prev = ids.pop(0)
     continue_list = [prev, prev]
@@ -167,10 +188,18 @@ def grouped_slice(records, count=None):
     if count is None:
         count = Transaction().database.IN_MAX
     count = max(1, count)
+    if not isinstance(records, Sized):
+        records = list(records)
     for i in range(0, len(records), count):
         yield islice(records, i, i + count)
 
 
+def pairwise_longest(iterable):
+    a, b = tee(iterable)
+    next(b, None)
+    return zip_longest(a, b)
+
+
 def is_instance_method(cls, method):
     for klass in cls.__mro__:
         type_ = klass.__dict__.get(method)
@@ -200,19 +229,18 @@ def strip_wildcard(string, wildcard='%', escape='\\'):
 
 def lstrip_wildcard(string, wildcard='%'):
     "Strip starting wildcard from string"
-    if not string:
-        return string
-    return string.lstrip(wildcard)
+    if string and string.startswith(wildcard):
+        string = string[1:]
+    return string
 
 
 def rstrip_wildcard(string, wildcard='%', escape='\\'):
     "Strip ending wildcard from string"
-    if not string:
-        return string
-    new_string = string.rstrip(wildcard)
-    if new_string[-1] == escape:
-        return string
-    return new_string
+    if (string
+            and string.endswith(wildcard)
+            and not string.endswith(escape + wildcard)):
+        string = string[:-1]
+    return string
 
 
 def escape_wildcard(string, wildcards='%_', escape='\\'):
@@ -228,13 +256,23 @@ def unescape_wildcard(string, wildcards='%_', escape='\\'):
 
 
 def is_full_text(value, escape='\\'):
-    escaped = value.strip('%')
+    escaped = strip_wildcard(value, escape=escape)
     escaped = escaped.replace(escape + '%', '').replace(escape + '_', '')
     if '%' in escaped or '_' in escaped:
         return False
     return value.startswith('%') == value.endswith('%')
 
 
+def likify(string, escape='\\'):
+    if not string:
+        return '%'
+    escaped = string.replace(escape + '%', '').replace(escape + '_', '')
+    if '%' in escaped or '_' in escaped:
+        return string
+    else:
+        return '%' + string + '%'
+
+
 _slugify_strip_re = re.compile(r'[^\w\s-]')
 _slugify_hyphenate_re = re.compile(r'[-\s]+')
 
@@ -272,3 +310,13 @@ def firstline(text):
         return next((x for x in text.splitlines() if x.strip()))
     except StopIteration:
         return ''
+
+
+def remove_forbidden_chars(value):
+    from trytond.model.fields import Char
+    if value is None:
+        return value
+    for c in Char.forbidden_chars:
+        if c in value:
+            value = value.replace(c, ' ')
+    return value.strip()
